"""Example of NaCl calculation."""

import numpy as np

from phonopy import Phonopy
from phonopy.file_IO import parse_BORN, parse_FORCE_SETS
from phonopy.interface.vasp import read_vasp

# from phonopy.structure.atoms import PhonopyAtoms


def _append_band(bands, q_start, q_end):
    band = []
    for i in range(51):
        band.append(np.array(q_start) + (np.array(q_end) - np.array(q_start)) / 50 * i)
    bands.append(band)


# NaCl crystal structure is read from POSCAR.
unitcell = read_vasp("POSCAR-unitcell")
# This can be given via a PhonopyAtoms class as follows:
# unitcell = PhonopyAtoms(symbols=(['Na'] * 4 + ['Cl'] * 4),
#                         cell=(np.eye(3) * 5.6903014761756712),
#                         scaled_positions=[[0, 0, 0],
#                                           [0, 0.5, 0.5],
#                                           [0.5, 0, 0.5],
#                                           [0.5, 0.5, 0],
#                                           [0.5, 0.5, 0.5],
#                                           [0.5, 0, 0],
#                                           [0, 0.5, 0],
#                                           [0, 0, 0.5]])

phonon = Phonopy(
    unitcell,
    [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
)

symmetry = phonon.symmetry
print("Space group: %s" % symmetry.get_international_table())

force_sets = parse_FORCE_SETS()
phonon.dataset = force_sets
phonon.produce_force_constants()
primitive = phonon.primitive

# Born effective charges and dielectric constants are read from BORN file.
nac_params = parse_BORN(primitive, filename="BORN")
# Or it can be of course given by hand as follows:
# born = [[[1.08703, 0, 0],
#          [0, 1.08703, 0],
#          [0, 0, 1.08703]],
#         [[-1.08672, 0, 0],
#          [0, -1.08672, 0],
#          [0, 0, -1.08672]]]
# epsilon = [[2.43533967, 0, 0],
#            [0, 2.43533967, 0],
#            [0, 0, 2.43533967]]
# factors = 14.400
# nac_params = {'born': born,
#               'factor': factors,
#               'dielectric': epsilon}
phonon.nac_params = nac_params

# BAND = 0.0 0.0 0.0  0.5 0.0 0.0  0.5 0.5 0.0  0.0 0.0 0.0  0.5 0.5 0.5
bands = []
_append_band(bands, [0.0, 0.0, 0.0], [0.5, 0.0, 0.0])
_append_band(bands, [0.5, 0.0, 0.0], [0.5, 0.5, 0.0])
_append_band(bands, [0.5, 0.5, 0.0], [0.0, 0.0, 0.0])
_append_band(bands, [0.0, 0.0, 0.0], [0.5, 0.5, 0.5])
phonon.run_band_structure(bands)
band_dict = phonon.get_band_structure_dict()
q_points = band_dict["qpoints"]
distances = band_dict["distances"]
frequencies = band_dict["frequencies"]
eigvecs = band_dict["eigenvectors"]
for q_path, d_path, freq_path in zip(q_points, distances, frequencies):
    for q, d, freq in zip(q_path, d_path, freq_path):
        print(
            ("%10.5f  %5.2f %5.2f %5.2f " + (" %7.3f" * len(freq)))
            % ((d, q[0], q[1], q[2]) + tuple(freq))
        )

phonon.plot_band_structure().show()

# Mesh sampling 20x20x20
phonon.run_mesh(mesh=[20, 20, 20])
phonon.run_thermal_properties(t_step=10, t_max=1000, t_min=0)

# DOS
phonon.run_total_dos(sigma=0.1)
dos_dict = phonon.get_total_dos_dict()
for omega, dos in zip(dos_dict["frequency_points"], dos_dict["total_dos"]):
    print("%15.7f%15.7f" % (omega, dos))
phonon.plot_total_dos().show()

# Thermal properties
tprop_dict = phonon.get_thermal_properties_dict()

for t, free_energy, entropy, cv in zip(
    tprop_dict["temperatures"],
    tprop_dict["free_energy"],
    tprop_dict["entropy"],
    tprop_dict["heat_capacity"],
):
    print(("%12.3f " + "%15.7f" * 3) % (t, free_energy, entropy, cv))
phonon.plot_thermal_properties().show()

# PDOS
phonon.run_mesh(mesh=[10, 10, 10], is_mesh_symmetry=False, with_eigenvectors=True)
phonon.run_projected_dos(use_tetrahedron_method=True)
pdos_dict = phonon.get_projected_dos_dict()
omegas = pdos_dict["frequency_points"]
pdos = pdos_dict["projected_dos"]
pdos_indices = [[0], [1]]
phonon.plot_projected_dos(pdos_indices=pdos_indices, legend=pdos_indices).show()



"""Calculate irreducible representation from eigenvectors."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from typing import Union

import numpy as np

from phonopy.harmonic.derivative_dynmat import DerivativeOfDynamicalMatrix
from phonopy.harmonic.dynamical_matrix import DynamicalMatrix, DynamicalMatrixNAC
from phonopy.phonon.character_table import character_table
from phonopy.phonon.degeneracy import degenerate_sets as get_degenerate_sets
from phonopy.physical_units import get_physical_units
from phonopy.structure.cells import is_primitive_cell
from phonopy.structure.symmetry import Symmetry
from phonopy.utils import similarity_transformation


class IrReps:
    """Class to calculate irreducible representations from eigenvectors.

    Methods and terminologies used in this class may be easily found
    in textbooks such as
    - Group theory with applications in chemical physics by Patrick Jacobs
    - Symmetry and condensed matter physics by M. El-Batanouny and F. Wooten

    """

    def __init__(
        self,
        dynamical_matrix: Union[DynamicalMatrix, DynamicalMatrixNAC],
        q,
        is_little_cogroup=False,
        nac_q_direction=None,
        factor=None,
        symprec=1e-5,
        degeneracy_tolerance=None,
        log_level=0,
    ):
        """Init method."""
        self._is_little_cogroup = is_little_cogroup
        self._nac_q_direction = nac_q_direction
        if factor is None:
            self._factor = get_physical_units().DefaultToTHz
        else:
            self._factor = factor
        self._log_level = log_level

        self._q = np.array(q)
        if degeneracy_tolerance is None:
            self._degeneracy_tolerance = 1e-5
        else:
            self._degeneracy_tolerance = degeneracy_tolerance
        self._symprec = symprec
        self._primitive = dynamical_matrix.primitive
        self._dynamical_matrix = dynamical_matrix
        self._ddm = DerivativeOfDynamicalMatrix(dynamical_matrix)
        self._character_table = None

        self._symmetry_dataset = Symmetry(
            self._primitive, symprec=self._symprec
        ).dataset

        if not is_primitive_cell(self._symmetry_dataset.rotations):
            raise RuntimeError(
                "Non-primitve cell is used. Your unit cell may be transformed to "
                "a primitive cell by PRIMITIVE_AXIS tag."
            )

    def run(self):
        """Calculate irreps."""
        self._set_eigenvectors(self._dynamical_matrix)

        (self._rotations_at_q, self._translations_at_q) = self._get_rotations_at_q()

        self._g = len(self._rotations_at_q)

        (
            self._pointgroup_symbol,
            self._transformation_matrix,
            self._conventional_rotations,
        ) = self._get_conventional_rotations()

        self._ground_matrices = self._get_ground_matrix()
        self._degenerate_sets = self._get_degenerate_sets()
        self._irreps = self._get_irreps()
        self._characters, self._irrep_dims = self._get_characters()

        self._ir_labels = None

        if (
            self._pointgroup_symbol in character_table.keys()
            and character_table[self._pointgroup_symbol] is not None
        ):
            self._rotation_symbols = self._get_rotation_symbols()
            if (abs(self._q) < self._symprec).all() and self._rotation_symbols:
                self._ir_labels = self._get_ir_labels()
            elif (abs(self._q) < self._symprec).all():
                if self._log_level > 0:
                    print("Database for this point group is not preprared.")
            else:
                if self._log_level > 0:
                    print("Database for non-Gamma point is not prepared.")
        else:
            self._rotation_symbols = None

        return True

    def _get_degenerate_sets(self):
        deg_sets = get_degenerate_sets(self._freqs, cutoff=self._degeneracy_tolerance)
        self._ddm.run(self._q)
        return deg_sets

    @property
    def band_indices(self):
        """Return band indices.

        Returns
        -------
        See docstring of ``degenerate_sets``.

        """
        return self._degenerate_sets

    def get_band_indices(self):
        """Return band indices."""
        warnings.warn(
            "IrReps.get_band_indices() is deprecated. "
            "Use IrReps.band_indices attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.band_indices

    @property
    def characters(self):
        """Return characters of irreps."""
        return self._characters

    def get_characters(self):
        """Return characters of irreps."""
        warnings.warn(
            "IrReps.get_characters() is deprecated. Use IrReps.characters attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.characters

    @property
    def eigenvectors(self):
        """Return eigenvectors."""
        return self._eig_vecs

    def get_eigenvectors(self):
        """Return eigenvectors."""
        warnings.warn(
            "IrReps.get_eigenvectors() is deprecated. "
            "Use IrReps.eigenvectors attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.eigenvectors

    @property
    def irreps(self):
        """Return irreps."""
        return self._irreps

    def get_irreps(self):
        """Return irreps."""
        warnings.warn(
            "IrReps.get_irreps() is deprecated. Use IrReps.irreps attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.irreps

    @property
    def ground_matrices(self):
        """Return ground matrices."""
        return self._ground_matrices

    def get_ground_matrices(self):
        """Return ground matrices."""
        warnings.warn(
            "IrReps.get_ground_matrices() is deprecated. "
            "Use IrReps.ground_matrices attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.ground_matrices

    @property
    def rotation_symbols(self):
        """Return symbols assigned to rotation matrices."""
        return self._rotation_symbols

    def get_rotation_symbols(self):
        """Return symbols assigned to rotation matrices."""
        warnings.warn(
            "IrReps.get_rotation_symbols() is deprecated. "
            "Use IrReps.rotation_symbols attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.rotation_symbols

    @property
    def conventional_rotations(self):
        """Return rotation matrices."""
        return self._conventional_rotations

    def get_rotations(self):
        """Return rotation matrices."""
        warnings.warn(
            "IrReps.get_rotations() is deprecated. "
            "Use IrReps.conventional_rotations attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.conventional_rotations

    def get_projection_operators(self, idx_irrep, i=None, j=None):
        """Return projection operators."""
        if i is None or j is None:
            return self._get_character_projection_operators(idx_irrep)
        else:
            return self._get_projection_operators(idx_irrep, i, j)

    @property
    def qpoint(self):
        """Return q-point."""
        return self._q

    def show(self, show_irreps=False):
        """Show irreps."""
        self._show(show_irreps)

    def write_yaml(self, show_irreps=False):
        """Write irreps in yaml file."""
        self._write_yaml(show_irreps)

    def _set_eigenvectors(self, dm):
        if self._nac_q_direction is not None and (np.abs(self._q) < 1e-5).all():
            dm.run(self._q, q_direction=self._nac_q_direction)
        else:
            dm.run(self._q)
        eig_vals, self._eig_vecs = np.linalg.eigh(dm.dynamical_matrix)
        self._freqs = np.sqrt(abs(eig_vals)) * np.sign(eig_vals) * self._factor

    def _get_rotations_at_q(self):
        rotations_at_q = []
        trans_at_q = []
        for r, t in zip(
            self._symmetry_dataset.rotations, self._symmetry_dataset.translations
        ):
            diff = np.dot(self._q, r) - self._q
            if (abs(diff - np.rint(diff)) < self._symprec).all():
                rotations_at_q.append(r)
                for i in range(3):
                    if np.abs(t[i] - 1) < self._symprec:
                        t[i] = 0.0
                trans_at_q.append(t)

        return np.array(rotations_at_q), np.array(trans_at_q)

    def _get_conventional_rotations(self):
        rotations = self._rotations_at_q.copy()
        pointgroup_symbol = self._symmetry_dataset.pointgroup
        transformation_matrix = self._symmetry_dataset.transformation_matrix
        conventional_rotations = self._transform_rotations(
            transformation_matrix, rotations
        )

        return (pointgroup_symbol, transformation_matrix, conventional_rotations)

    def _transform_rotations(self, tmat, rotations):
        trans_rots = []

        for r in rotations:
            r_conv = similarity_transformation(tmat, r)
            trans_rots.append(np.rint(r_conv).astype(int))

        return np.array(trans_rots)

    def _get_ground_matrix(self):
        matrices = []

        for r, t in zip(self._rotations_at_q, self._translations_at_q):
            lat = self._primitive.cell.T
            r_cart = similarity_transformation(lat, r)
            perm_mat = self._get_modified_permutation_matrix(r, t)
            matrices.append(np.kron(perm_mat, r_cart))

        return np.array(matrices)

    def _get_characters(self):
        characters = []
        irrep_dims = []
        for irrep_Rs in self._irreps:
            characters.append([np.trace(rep) for rep in irrep_Rs])
            irrep_dims.append(len(irrep_Rs[0]))
        return np.array(characters), np.array(irrep_dims)

    def _get_modified_permutation_matrix(self, r, t):
        num_atom = len(self._primitive)
        pos = self._primitive.scaled_positions
        matrix = np.zeros((num_atom, num_atom), dtype=complex)
        for i, p1 in enumerate(pos):
            p_rot = np.dot(r, p1) + t  # i -> j
            for j, p2 in enumerate(pos):
                diff = p_rot - p2  # Rx_i + t - x_j
                if (abs(diff - np.rint(diff)) < self._symprec).all():
                    phase_factor = np.dot(
                        self._q, np.dot(np.linalg.inv(r), p2 - t) - p2
                    )
                    if self._is_little_cogroup:
                        phase_factor = np.dot(t, self._q)
                    matrix[j, i] = np.exp(2j * np.pi * phase_factor)
        return matrix

    def _get_irreps(self):
        eigvecs = self._eig_vecs.T
        irrep = []
        for band_indices in self._degenerate_sets:
            irrep_Rs = []
            for mat in self._ground_matrices:
                n_deg = len(band_indices)

                if n_deg == 1:
                    vec = eigvecs[band_indices[0]]
                    irrep_Rs.append([[np.vdot(vec, np.dot(mat, vec))]])
                    continue

                irrep_R = np.zeros((n_deg, n_deg), dtype=complex)
                for i, b_i in enumerate(band_indices):
                    vec_i = eigvecs[b_i]
                    for j, b_j in enumerate(band_indices):
                        vec_j = eigvecs[b_j]
                        irrep_R[i, j] = np.vdot(vec_i, np.dot(mat, vec_j))
                irrep_Rs.append(irrep_R)

            irrep.append(irrep_Rs)

        return irrep

    def _get_character_projection_operators(self, idx_irrep):
        dim = self._irrep_dims[idx_irrep]
        chars = self._characters[idx_irrep]
        return (
            np.sum(
                [mat * char.conj() for mat, char in zip(self._ground_matrices, chars)],
                axis=0,
            )
            * dim
            / self._g
        )

    def _get_projection_operators(self, idx_irrep, i, j):
        dim = self._irrep_dims[idx_irrep]
        return (
            np.sum(
                [
                    mat * r[i, j].conj()
                    for mat, r in zip(self._ground_matrices, self._irreps[idx_irrep])
                ],
                axis=0,
            )
            * dim
            / self._g
        )

    def _get_rotation_symbols(self):
        ptg_symbol = self._pointgroup_symbol
        for ct in character_table[ptg_symbol]:
            mapping_table = ct["mapping_table"]
            rotation_symbols = []
            for r in self._conventional_rotations:
                rotation_symbols.append(_get_rotation_symbol(r, mapping_table))
            if False in rotation_symbols:
                ret_val = None
            else:
                ret_val = rotation_symbols
            if ret_val is not None:
                self._character_table = ct
                break

        return ret_val

    def _get_ir_labels(self):
        ir_labels = []
        rot_list = self._character_table["rotation_list"]
        char_table = self._character_table["character_table"]
        for chars in self._characters:
            chars_ordered = np.zeros(len(rot_list), dtype=complex)
            for rs, ch in zip(self._rotation_symbols, chars):
                chars_ordered[rot_list.index(rs)] += ch

            for i, rl in enumerate(rot_list):
                chars_ordered[i] /= len(self._character_table["mapping_table"][rl])

            found = False
            for ct_label in char_table.keys():
                if (
                    abs(chars_ordered - np.array(char_table[ct_label])) < self._symprec
                ).all():
                    ir_labels.append(ct_label)
                    found = True
                    break

            if not found:
                ir_labels.append(None)

            if self._log_level > 1:
                text = ""
                for v in chars_ordered:
                    text += "%5.2f " % abs(v)
                if found:
                    print("%s %s" % (text, ct_label))
                else:
                    print("%s Not found" % text)

        return ir_labels

    def _show(self, show_irreps):
        print("")
        print("-------------------------------")
        print("  Irreducible representations")
        print("-------------------------------")
        print("q-point: %s" % self._q)
        print("Point group: %s" % self._pointgroup_symbol)
        print("")

        if (np.abs(self._q) < self._symprec).all():
            width = 6
            print("Original rotation matrices:")
            print("")
            _print_rotations(self._rotations_at_q, width=width)
        else:
            width = 4
            print("Original symmetry operations:")
            print("")
            _print_rotations(
                self._rotations_at_q, translations=self._translations_at_q, width=width
            )

        print("Transformation matrix:")
        print("")
        for v in self._transformation_matrix:
            print("%6.3f %6.3f %6.3f" % tuple(v))
        print("")
        print("Rotation matrices by transformation matrix:")
        print("")
        _print_rotations(
            self._conventional_rotations,
            rotation_symbols=self._rotation_symbols,
            width=width,
        )
        print("Character table:")
        print("")
        for i, deg_set in enumerate(self._degenerate_sets):
            text = "%3d (%8.3f): " % (deg_set[0] + 1, self._freqs[deg_set[0]])
            if self._ir_labels is None:
                print(text)
            elif self._ir_labels[i] is None:
                warning = "Not found. Try adjusting tolerance value in IRREPS."
                print("%s%s" % (text, warning))
            else:
                print("%s%s" % (text, self._ir_labels[i]))
            _print_characters(self._characters[i])
            print("")

        if show_irreps:
            self._show_irreps()

    def _show_irreps(self):
        print("IR representations:")
        print("")

        for deg_set, irrep_Rs in zip(self._degenerate_sets, self._irreps):
            print("%3d (%8.3f):" % (deg_set[0] + 1, self._freqs[deg_set[0]]))
            print("")
            for j, irrep_R in enumerate(irrep_Rs):
                for k, irrep_Rk in enumerate(irrep_R):
                    text = "     "
                    for ll, irrep_Rkl in enumerate(irrep_Rk):
                        if irrep_Rkl.real > 0:
                            sign_r = " "
                        else:
                            sign_r = "-"

                        if irrep_Rkl.imag > 0:
                            sign_i = "+"
                        else:
                            sign_i = "-"

                        if k == 0:
                            str_index = "%2d" % (j + 1)
                        else:
                            str_index = "  "

                        if ll > 0:
                            str_index = ""

                        text += "%s (%s%5.3f %s%5.3fi) " % (
                            str_index,
                            sign_r,
                            abs(irrep_Rkl.real),
                            sign_i,
                            abs(irrep_Rkl.imag),
                        )
                    print(text)
                if len(irrep_R) > 1:
                    print("")
            if len(irrep_R) == 1:
                print("")

    def _write_yaml(self, show_irreps):
        w = open("irreps.yaml", "w")
        w.write("q-position: [ %12.7f, %12.7f, %12.7f ]\n" % tuple(self._q))
        w.write("point_group: %s\n" % self._pointgroup_symbol)
        w.write("transformation_matrix:\n")
        for v in self._transformation_matrix:
            w.write("- [ %10.7f, %10.7f, %10.7f ]\n" % tuple(v))
        w.write("rotations:\n")
        for i, r in enumerate(self._conventional_rotations):
            w.write("- matrix:\n")
            for v in r:
                w.write("  - [ %2d, %2d, %2d ]\n" % tuple(v))
            if self._rotation_symbols:
                w.write("  symbol: %s\n" % self._rotation_symbols[i])
        w.write("normal_modes:\n")
        for i, deg_set in enumerate(self._degenerate_sets):
            w.write("- band_indices: [ ")
            w.write("%d" % (deg_set[0] + 1))
            for bi in deg_set[1:]:
                w.write(", %d" % (bi + 1))
            w.write(" ]\n")
            w.write("  frequency: %-15.10f\n" % self._freqs[deg_set[0]])
            if self._ir_labels:
                w.write("  ir_label: %s\n" % self._ir_labels[i])
            w.write("  characters: ")
            chars = np.rint(np.abs(self._characters[i]))
            phase = (np.angle(self._characters[i]) / np.pi * 180) % 360
            if len(chars) > 1:
                w.write("[ [ %2d, %5.1f ]" % (chars[0], phase[0]))
                for chi, theta in zip(chars[1:], phase[1:]):
                    w.write(", [ %2d, %5.1f ]" % (chi, theta))
                w.write(" ]\n")
            else:
                w.write("[ [ %2d, %5.1f ] ]\n" % (chars[0], phase[0]))

        if show_irreps:
            self._write_yaml_irreps(w)

        w.close()

    def _write_yaml_irreps(self, file_pointer):
        w = file_pointer
        if not self._irreps:
            self._irrep = self._get_irreps()

        w.write("\n")
        w.write("irreps:\n")
        for i, (deg_set, irrep_Rs) in enumerate(
            zip(self._degenerate_sets, self._irreps)
        ):
            w.write("- # %d\n" % (i + 1))
            for j, irrep_R in enumerate(irrep_Rs):
                if self._rotation_symbols:
                    symbol = self._rotation_symbols[j]
                else:
                    symbol = ""
                if len(deg_set) > 1:
                    w.write("  - # %d %s\n" % (j + 1, symbol))
                    for _, v in enumerate(irrep_R):
                        w.write("    - [ ")
                        for x in v[:-1]:
                            w.write("%10.7f, %10.7f,   " % (x.real, x.imag))
                        w.write("%10.7f, %10.7f ] # (" % (v[-1].real, v[-1].imag))

                        w.write(
                            ("%5.0f" * len(v))
                            % tuple((np.angle(v) / np.pi * 180) % 360)
                        )
                        w.write(")\n")
                else:
                    x = irrep_R[0][0]
                    w.write(
                        "  - [ [ %10.7f, %10.7f ] ] # (%3.0f) %d %s\n"
                        % (
                            x.real,
                            x.imag,
                            (np.angle(x) / np.pi * 180) % 360,
                            j + 1,
                            symbol,
                        )
                    )

        pass


def _get_rotation_symbol(rotation, mapping_table):
    for k in mapping_table:
        v = mapping_table[k]
        for r in v:
            if (r == rotation).all():
                return k
    return False


def _print_characters(characters, width=6):
    text = ""
    for i, c in enumerate(characters):
        angle = np.angle(c) / np.pi * 180
        if angle < 0:
            angle += 360
        angle = np.around(angle, decimals=2) % 360
        val = abs(c)
        if val < 1e-5:
            val = 0
            angle = 0
        else:
            val = np.rint(val)
        text += "(%2d, %5.1f) " % (val, angle)
        if ((i + 1) % width == 0 and i != 0) or (len(characters) == i + 1):
            print("    " + text)
            text = ""


def _get_rotation_text(rotations, translations, rotation_symbols, width, num_rest, i):
    lines = []
    if rotation_symbols is None:
        if translations is None:
            lines.append(
                ("    %2d    " * num_rest)
                % tuple(np.arange(i * width, i * width + num_rest) + 1)
            )
        else:
            lines.append(
                ("       %2d       " * num_rest)
                % tuple(np.arange(i * width, i * width + num_rest) + 1)
            )
    else:
        text = ""
        for k in range(num_rest):
            rot_symbol = rotation_symbols[i * width + k]
            if translations is None:
                if len(rot_symbol) < 3:
                    text += "    %2s    " % rot_symbol
                elif len(rot_symbol) == 3:
                    text += "    %3s   " % rot_symbol
                elif len(rot_symbol) == 4:
                    text += "   %4s   " % rot_symbol
                else:
                    text += "   %5s  " % rot_symbol
            else:
                if len(rot_symbol) < 3:
                    text += "       %2s        " % rot_symbol
                else:
                    text += "     %5s     " % rot_symbol
        lines.append(text)
    if translations is None:
        lines.append(" -------- " * num_rest)
    else:
        lines.append(" -------------- " * num_rest)

    for j in range(3):
        text = ""
        for k in range(num_rest):
            text += " %2d %2d %2d " % tuple(rotations[i * width + k][j])
            if translations is not None:
                text += "%5.2f " % translations[i * width + k][j]
        lines.append(text)
    lines.append("")

    return "\n".join(lines)


def _print_rotations(rotations, translations=None, rotation_symbols=None, width=6):
    for i in range(len(rotations) // width):
        print(
            _get_rotation_text(
                rotations, translations, rotation_symbols, width, width, i
            )
        )

    num_rest = len(rotations) % width
    if num_rest > 0:
        i = len(rotations) // width
        print(
            _get_rotation_text(
                rotations, translations, rotation_symbols, width, num_rest, i
            )
        )



"""Transform dynamical matrix to force constants."""

# Copyright (C) 2014 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from typing import TYPE_CHECKING

import numpy as np

from phonopy.harmonic.force_constants import distribute_force_constants_by_translations
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive, get_supercell, sparse_to_dense_svecs
from phonopy.structure.snf import SNF3x3

if TYPE_CHECKING:
    from phonopy import Phonopy


def get_commensurate_points(supercell_matrix):  # wrt primitive cell
    """Commensurate q-points are returned.

    Parameters
    ----------
    supercell_matrix : array_like
        Supercell matrix with respect to primitive cell basis vectors.
        shape=(3, 3), dtype=int

    Returns
    -------
    commensurate_points : ndarray
        Commensurate points corresponding to supercell matrix.
        shape=(N, 3), dtype='double', order='C'
        where N = det(supercell_matrix)

    """
    smat = np.array(supercell_matrix, dtype=int)
    rec_primitive = PhonopyAtoms(
        symbols=["H"], scaled_positions=[[0, 0, 0]], cell=np.diag([1, 1, 1])
    )
    rec_supercell = get_supercell(rec_primitive, smat.T)
    q_pos = rec_supercell.scaled_positions
    return np.array(
        np.where(q_pos > 1 - 1e-15, q_pos - 1, q_pos), dtype="double", order="C"
    )


def get_commensurate_points_in_integers(supercell_matrix):
    """Commensurate q-points in integer representation are returned.

    A set of integer representation of lattice points is transformed to
    the equivalent set of lattice points in fractional coordinates with
    respect to supercell basis vectors by
        integer_lattice_points / det(supercell_matrix)

    Parameters
    ----------
    supercell_matrix : array_like
        Supercell matrix with respect to primitive cell basis vectors.
        shape=(3, 3)
        dtype=intc

    Returns
    -------
    lattice_points : ndarray
        Integer representation of lattice points in supercell.
        shape=(N, 3)

    """
    smat = np.array(supercell_matrix, dtype=int)
    snf = SNF3x3(smat.T)
    snf.run()
    D = snf.D.diagonal()
    b, c, a = np.meshgrid(range(D[1]), range(D[2]), range(D[0]))
    lattice_points = np.dot(
        np.c_[
            a.ravel() * D[1] * D[2], b.ravel() * D[0] * D[2], c.ravel() * D[0] * D[1]
        ],
        snf.Q.T,
    )
    lattice_points = np.array(lattice_points % np.prod(D), dtype="intc", order="C")
    return lattice_points


def categorize_commensurate_points(comm_points):
    """Categorize integer commensurate points.

    Points are sorted by either q = -q + G or q != -q + G.

    """
    N = len(comm_points)
    ii = []
    ij = []
    for i, p in enumerate(comm_points):
        for j, _p in enumerate(comm_points):
            if ((p + _p) % N == 0).all():
                if i == j:
                    ii.append(i)
                elif i < j:
                    ij.append(i)
                break

    assert len(ii) + len(ij) * 2 == len(comm_points)

    return ii, ij


def ph2fc(ph_orig: "Phonopy", supercell_matrix, with_nac=True):
    """Transform force constants in Phonopy instance to other shape.

    This function is deprecated. Use ph2ph or Phonopy.ph2ph.

    Parameters
    ----------
    supercell_matrix : array_like
        This specifies array shape of the force constants.
    with_nac : bool, optional
        Use non-analytical term correction if NAC paramerters exist. Default is
        True.

    Returns
    -------
    force_constants : ndarray
        Transformed force constants of ``supercell_matrix``.

    """
    warnings.warn(
        "ph2fc function is deprecated. Use Phonopy.ph2ph instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return ph2ph(ph_orig, supercell_matrix, with_nac=with_nac).force_constants


def ph2ph(ph_orig: "Phonopy", supercell_matrix, with_nac=False) -> "Phonopy":
    """Transform force constants in Phonopy instance to other shape.

    Parameters
    ----------
    supercell_matrix : array_like
        This specifies array shape of the force constants.
    with_nac : bool, optional
        Non-analytical term correction (NAC) is used under the Fourier
        interpolation and NAC parameters are copied to Phonopy class
        instance if they exist. Default is False.

    Returns
    -------
    ph : Phonopy
        Phonopy class instance with init parameters of this Phonopy class
        instance and transformed force constants of `supercell_matrix`.

    """
    return ph_orig.ph2ph(supercell_matrix, with_nac=with_nac)


class DynmatToForceConstants:
    """Transforms eigensolutions to force constants.

    This is the inverse transform of force constants to eigensolutions.
    Eigenvalue-eigenvector pairs of dynamical matrices are transformed to
    force constants by this class. Optionally, dynamical matrices
    (not the DynamicalMatrix class instance, but the calculated matrices)
    can be used instead of above pairs.

    The eigensolutions have to be those calculated at commensurate q-points
    of supercell matrix. The commensurate q-points are obtained by the
    attribute ``commensurate_points``. This set of commensurate points can be
    also given by setter of it.

    One supposed usage is::

       d2f = DynmatToForceConstants(primitive, supercell)
       comm_points = d2f.commensurate_points
       ... calculated phonons at comm_points
       d2f.create_dynamical_matrices(eigenvalues=eigenvalues,
                                     eigenvectors=eigenvectors)
       d2f.run()
       fc = d2f.force_constants

    Instead of recreating dynamical matrices from eigensolutions,
    dynamical matrices can be used directly as follows::

       d2f = DynmatToForceConstants(primitive, supercell)
       comm_points = d2f.commensurate_points
       ... calculated phonons at comm_points
       d2f.dynamical_matrices = dynmat
       d2f.run()
       fc = d2f.force_constants

    Attributes
    ----------
    force_constants : ndarray
        Calculated force constants. Only getter. Array shape depends on
        ``is_full_fc``.
            True : shape=(supercell_atoms, supercell_atoms, 3, 3)
            False : shape=(primitive_atoms, supercell_atoms, 3, 3)
        Default is True.
        dtype='double', order='C'.
    commensurate_points : ndarray
        Commensurate q-points of supercell matrix.
        shape=(det(supercell_matrix), 3), dtype='double', order='C'.
    dynamical_matrices : ndarray
        shape=(det(supercell_matrix), num_band, num_band),
        where num_band is 3 x number of atoms in primitive cell.
        dtype=complex of "c%d" % (np.dtype('double').itemsize * 2)
        order='C'.

    """

    def __init__(
        self,
        primitive: Primitive,
        supercell: PhonopyAtoms,
        eigenvalues=None,
        eigenvectors=None,
        dynamical_matrices=None,
        commensurate_points=None,
        is_full_fc=True,
        use_openmp=False,
    ):
        """Init method.

        Parameters
        ----------
        primitive : Primitive
            Primitive cell
        supercell : PhonopyAtoms
            Supercell, not necessarily being an instance of Supercell class.
        is_full_fc : bool
            This controls the matrix shape of calculated force constants.
            True and False give the full and compact force cosntants,
            respectively. The default is True. See more details in Attributes
            section of this class.
        use_openmp : bool, optional, default=False
            Use OpenMP in calculate force constants from dynamical matrix.

        """
        self._pcell = primitive
        self._scell = supercell
        supercell_matrix = np.linalg.inv(self._pcell.primitive_matrix)
        supercell_matrix = np.rint(supercell_matrix).astype("intc")
        if commensurate_points is None:
            self._commensurate_points = get_commensurate_points(supercell_matrix)
        else:
            self._commensurate_points = commensurate_points

        svecs, multi = self._pcell.get_smallest_vectors()
        if self._pcell.store_dense_svecs:
            self._svecs = svecs
            self._multi = multi
        else:
            self._svecs, self._multi = sparse_to_dense_svecs(svecs, multi)

        self._use_openmp = use_openmp

        self._dynmat = None
        self._fc = None

        n_s = len(self._scell)
        n_p = len(self._pcell)
        if is_full_fc:
            self._fc_shape = (n_s, n_s, 3, 3)
        else:
            self._fc_shape = (n_p, n_s, 3, 3)

        self._dtype_complex = "c%d" % (np.dtype("double").itemsize * 2)

        if dynamical_matrices is not None or commensurate_points is not None:
            warnings.warn(
                "Instanciation init parameters of dynamical_matrices"
                " and commensurate_points are deprecated. Use "
                "respecitve attributes.",
                DeprecationWarning,
                stacklevel=2,
            )

        if eigenvalues is not None or eigenvectors is not None:
            warnings.warn(
                "Instanciation init parameters of eigenvalues and "
                "eigenvectors are deprecated. Use "
                "create_dynamical_matrices method.",
                DeprecationWarning,
                stacklevel=2,
            )

        if eigenvalues is not None and eigenvectors is not None:
            self.create_dynamical_matrices(
                eigenvalues=eigenvalues, eigenvectors=eigenvectors
            )
        elif dynamical_matrices is not None:
            self.dynamical_matrices = dynamical_matrices

    def run(self, lang="C"):
        """Run."""
        self._fc = np.zeros(self._fc_shape, dtype="double", order="C")
        self._inverse_transformation(lang=lang)

    @property
    def force_constants(self):
        """Return force constants."""
        return self._fc

    def get_force_constants(self):
        """Return force constants."""
        warnings.warn(
            "Use attribute, force_constants.", DeprecationWarning, stacklevel=2
        )
        return self.force_constants

    @property
    def commensurate_points(self):
        """Getter and setter of commensurate points.

        Returns for getter and Parameters for setter
        --------------------------------------------
        commensurate_points : ndarray (array_like for setter)
            Commensurate q-points corresponding to supercell_matrix. The order
            has to be the same as those of eigenvalues and eigenvectors. As
            the default behaviour, commensurate q-points are generated unless
            they are given.
            shape=(det(supercell_matrix), 3), dtype='double', order='C'

        """
        return self._commensurate_points

    @commensurate_points.setter
    def commensurate_points(self, comm_points):
        self._commensurate_points = np.array(comm_points, dtype="double", order="C")

    def get_commensurate_points(self):
        """Commensurate points in supercell with respect to primitive cell."""
        warnings.warn(
            "Use attribute, commensurate_points.", DeprecationWarning, stacklevel=2
        )
        return self.commensurate_points

    @property
    def dynamical_matrices(self):
        """Getter and setter of numerical matrices of dynamical matrices.

        dynamical_matrices : ndarray (array_like for setter)
            Dynamical matrices at commensurate q-points.
            shape=(det(supercell_matrix), num_band, num_band)
            dtype=complex of "c%d" % (np.dtype('double').itemsize * 2)
            order='C'.
            where ``num_band`` is 3 x number of atoms in primitive cell.

        """
        return self._dynmat

    @dynamical_matrices.setter
    def dynamical_matrices(self, dynmat):
        self._dynmat = np.array(dynmat, dtype=self._dtype_complex, order="C")

    def get_dynamical_matrices(self):
        """Return numerical matrices of dynamical matrices."""
        warnings.warn(
            "Use attribute, dynamical_matrix.", DeprecationWarning, stacklevel=2
        )
        return self.dynamical_matrices

    def set_dynamical_matrices(self, dynmat):
        """Set numerical matrices of dynamical matrices."""
        warnings.warn(
            "Use attribute, dynamical_matrix.", DeprecationWarning, stacklevel=2
        )
        self.dynamical_matrices = dynmat

    def create_dynamical_matrices(self, eigenvalues, eigenvectors):
        """Create dynamcial matrices from eigenvalues and eigenvectors pairs.

        Parameters
        ----------
        eigenvalues : ndarray
            Phonon eigenvalues as the solution of dynamical matrices at
            commensurate q-points.
            shape=(det(supercell_matrix), num_band), dtype='double', order='C'
            where ``num_band`` is 3 x number of atoms in primitive cell.
        eigenvectors : ndarray
            Phonon eigenvectors as the solution of dynamical matrices at
            commensurate q-points.
            shape=(det(supercell_matrix), num_band, num_band)
            dtype=complex of "c%d" % (np.dtype('double').itemsize * 2)
            order='C'.
            where ``num_band`` is 3 x number of atoms in primitive cell.

        """
        dm = []
        for eigvals, eigvecs in zip(eigenvalues, eigenvectors):
            dm.append(np.dot(np.dot(eigvecs, np.diag(eigvals)), eigvecs.T.conj()))
        self.dynamical_matrices = dm

    def _inverse_transformation(self, lang="C"):
        if lang == "C":
            self._c_inverse_transformation()
        else:
            self._py_inverse_transformation()

        if self._fc.shape[0] == self._fc.shape[1]:
            distribute_force_constants_by_translations(self._fc, self._pcell)

    def _c_inverse_transformation(self):
        import phonopy._phonopy as phonoc

        s2p = np.array(self._pcell.s2p_map, dtype="int64")
        p2p = self._pcell.p2p_map
        s2pp = np.array([p2p[i] for i in s2p], dtype="int64")

        if self._fc.shape[0] == self._fc.shape[1]:
            fc_index_map = np.array(self._pcell.p2s_map, dtype="int64")
        else:
            fc_index_map = np.arange(self._fc.shape[0], dtype="int64")

        phonoc.transform_dynmat_to_fc(
            self._fc,
            self._dynmat.view(dtype="double"),
            self._commensurate_points,
            self._svecs,
            self._multi,
            self._pcell.masses,
            s2pp,
            fc_index_map,
            self._use_openmp * 1,
        )

    def _py_inverse_transformation(self):
        s2p = self._pcell.s2p_map
        p2s = self._pcell.p2s_map
        p2p = self._pcell.p2p_map

        m = self._pcell.masses
        N = len(self._scell) / len(self._pcell)

        for p_i, s_i in enumerate(p2s):
            for s_j, p_j in enumerate([p2p[i] for i in s2p]):
                coef = np.sqrt(m[p_i] * m[p_j]) / N
                fc_elem = self._sum_q(p_i, s_j, p_j) * coef
                if self._fc.shape[0] == self._fc.shape[1]:
                    self._fc[s_i, s_j] = fc_elem
                else:
                    self._fc[p_i, s_j] = fc_elem

    def _sum_q(self, p_i, s_j, p_j):
        """Sum over commensurate q-points for a pair of atoms."""
        multi, adrs = self._multi[s_j, p_i]
        pos = self._svecs[adrs : (adrs + multi)]
        sum_q = np.zeros((3, 3), dtype=self._dtype_complex, order="C")
        phases = -2j * np.pi * np.dot(self._commensurate_points, pos.T)
        phase_factors = np.exp(phases).sum(axis=1) / multi
        for i, coef in enumerate(phase_factors):
            sum_q += (
                self._dynmat[i, (p_i * 3) : (p_i * 3 + 3), (p_j * 3) : (p_j * 3 + 3)]
                * coef
            )
        return sum_q.real



"""Helper methods of phonopy loader."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import pathlib
from dataclasses import asdict
from typing import Literal, Optional, Union

import numpy as np

from phonopy import Phonopy
from phonopy.exception import ForcesetsNotFoundError
from phonopy.file_IO import (
    parse_BORN,
    parse_FORCE_CONSTANTS,
    parse_FORCE_SETS,
    read_force_constants_hdf5,
)
from phonopy.harmonic.force_constants import (
    compact_fc_to_full_fc,
    full_fc_to_compact_fc,
)
from phonopy.interface.calculator import (
    get_force_constant_conversion_factor,
    read_crystal_structure,
)
from phonopy.interface.pypolymlp import (
    PypolymlpParams,
    parse_mlp_params,
)
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive, get_primitive_matrix
from phonopy.structure.dataset import forces_in_dataset


def get_cell_settings(
    supercell_matrix=None,
    primitive_matrix=None,
    unitcell=None,
    supercell=None,
    unitcell_filename=None,
    supercell_filename=None,
    calculator=None,
    symprec=1e-5,
    log_level=0,
):
    """Return crystal structures."""
    optional_structure_info = None
    if primitive_matrix is None or (
        isinstance(primitive_matrix, str) and primitive_matrix == "auto"
    ):
        pmat = "auto"
    else:
        pmat = primitive_matrix

    if unitcell_filename is not None:
        cell, optional_structure_info = _read_crystal_structure(
            filename=unitcell_filename, interface_mode=calculator
        )
        smat = supercell_matrix
        if log_level:
            print(
                'Unit cell structure was read from "%s".' % optional_structure_info[0]
            )
    elif supercell_filename is not None:
        cell, optional_structure_info = read_crystal_structure(
            filename=supercell_filename, interface_mode=calculator
        )
        smat = np.eye(3, dtype="intc", order="C")
        if log_level:
            print(
                'Supercell structure was read from "%s".' % optional_structure_info[0]
            )
    elif unitcell is not None:
        cell = unitcell.copy()
        smat = supercell_matrix
    elif supercell is not None:
        cell = supercell.copy()
        smat = np.eye(3, dtype="intc", order="C")
    else:
        raise RuntimeError("Cell has to be specified.")

    if optional_structure_info is not None and cell is None:
        filename = optional_structure_info[0]
        msg = "'%s' could not be found." % filename
        raise FileNotFoundError(msg)

    pmat = get_primitive_matrix(pmat, symprec=symprec)

    return cell, smat, pmat


def get_nac_params(
    primitive: Optional[Primitive] = None,
    nac_params: Optional[dict] = None,
    born_filename: Optional[str] = None,
    is_nac: bool = True,
    nac_factor: Optional[float] = None,
    log_level: int = 0,
) -> dict:
    """Look for and return NAC parameters.

    Parameters
    ----------
    primitive : Primitive
        Primitive cell.
    nac_params : dict
        NAC parameters.
    born_filename : str
        Filename of BORN file.
    is_nac : bool
        Whether to read NAC parameters from BORN file.
    nac_factor : float
        Unit conversion factor for non-analytical term correction. This value is
        set only when the NAC factor is unset by other means.
    log_level : int
        Log level.

    Returns
    -------
    dict
        Parameters used for non-analytical term correction 'born': ndarray
            Born effective charges. shape=(primitive cell atoms, 3, 3),
            dtype='double', order='C'
        'dielectric': ndarray
            Dielectric constant tensor. shape=(3, 3), dtype='double', order='C'
        'factor': float, optional
            Unit conversion factor.

    """
    if born_filename is not None:
        _nac_params = parse_BORN(primitive, filename=born_filename)
        if log_level:
            print('NAC parameters were read from "%s".' % born_filename)
    elif nac_params is not None:  # nac_params input or phonopy_yaml.nac_params
        _nac_params = nac_params
    elif is_nac and pathlib.Path("BORN").exists():
        _nac_params = parse_BORN(primitive, filename="BORN")
        if log_level:
            print('NAC params were read from "BORN".')
    else:
        _nac_params = None

    if _nac_params and "factor" not in _nac_params and nac_factor is not None:
        _nac_params["factor"] = nac_factor

    return _nac_params


def read_force_constants_from_hdf5(
    filename="force_constants.hdf5", p2s_map=None, calculator=None
):
    """Convert force constants physical unit.

    Each calculator interface has own default force constants physical unit.
    This method reads 'physical_unit' in force constants hdf5 file and
    if this is different from the one for 'calculator', the force constants
    are converted to have the physical unit of the calculator.

    Note
    ----
    This method is also used from phonopy script.

    """
    fc, fc_unit = read_force_constants_hdf5(
        filename=filename, p2s_map=p2s_map, return_physical_unit=True
    )
    if fc_unit is None:
        return fc
    else:
        factor = get_force_constant_conversion_factor(fc_unit, calculator)
        return fc * factor


def select_and_load_dataset(
    phonon: Phonopy,
    dataset: Optional[dict],
    phonopy_yaml_filename: Optional[str] = None,
    force_sets_filename: Optional[str] = None,
    log_level: int = 0,
) -> Optional[dict]:
    """Set displacement-force dataset."""
    natom = len(phonon.supercell)
    _dataset = None
    _force_sets_filename = None
    if forces_in_dataset(dataset):
        _dataset = dataset
        _force_sets_filename = phonopy_yaml_filename
    elif force_sets_filename is not None:
        _dataset = parse_FORCE_SETS(natom=natom, filename=force_sets_filename)
        _force_sets_filename = force_sets_filename
    elif pathlib.Path("FORCE_SETS").exists():
        _dataset = parse_FORCE_SETS(natom=natom)
        _force_sets_filename = "FORCE_SETS"
    else:
        _dataset = dataset

    if log_level:
        if forces_in_dataset(_dataset):
            print(f'Displacement-force dataset was read from "{_force_sets_filename}".')
        elif _dataset is not None:
            print(f'Displacement dataset was read from "{_force_sets_filename}".')

    return _dataset


def select_and_extract_force_constants(
    phonon: Phonopy,
    phonopy_yaml_filename: Optional[str] = None,
    fc: Optional[np.ndarray] = None,  # From phonopy_yaml
    force_constants_filename: Optional[str] = None,
    is_compact_fc: bool = True,
    log_level: int = 0,
) -> Optional[np.ndarray]:
    """Set force constants."""
    _fc = None
    _force_constants_filename = None
    if fc is not None:
        _fc = fc
        _force_constants_filename = phonopy_yaml_filename
    elif force_constants_filename is not None:
        _fc = _read_force_constants_file(
            phonon,
            force_constants_filename,
            is_compact_fc=is_compact_fc,
            log_level=log_level,
        )
        _force_constants_filename = force_constants_filename
    elif phonon.force_constants is None:
        # unless provided these from phonopy_yaml.
        if pathlib.Path("FORCE_CONSTANTS").exists():
            _fc = _read_force_constants_file(
                phonon,
                "FORCE_CONSTANTS",
                is_compact_fc=is_compact_fc,
                log_level=log_level,
            )
            _force_constants_filename = "FORCE_CONSTANTS"
        elif pathlib.Path("force_constants.hdf5").exists():
            _fc = _read_force_constants_file(
                phonon,
                "force_constants.hdf5",
                is_compact_fc=is_compact_fc,
                log_level=log_level,
            )
            _force_constants_filename = "force_constants.hdf5"

    if _fc is not None:
        if log_level and _force_constants_filename is not None:
            print(f'Force constants were read from "{_force_constants_filename}".')
        if not is_compact_fc and _fc.shape[0] != _fc.shape[1]:
            _fc = compact_fc_to_full_fc(phonon.primitive, _fc, log_level=log_level)
        elif is_compact_fc and _fc.shape[0] == _fc.shape[1]:
            _fc = full_fc_to_compact_fc(phonon.primitive, _fc, log_level=log_level)

    return _fc


def prepare_pypolymlp_and_dataset(
    phonon: Phonopy,
    mlp_params: Optional[dict] = None,
    displacement_distance: Optional[float] = None,
    number_of_snapshots: Optional[Union[int, Literal["auto"]]] = None,
    random_seed: Optional[int] = None,
    prepare_dataset: bool = False,
    log_level: int = 0,
):
    """Prepare pypolymlp and dataset."""
    _run_pypolymlp(phonon, mlp_params, log_level=log_level)
    if prepare_dataset:
        _prepare_dataset_by_pypolymlp(
            phonon,
            displacement_distance=displacement_distance,
            number_of_snapshots=number_of_snapshots,
            random_seed=random_seed,
            log_level=log_level,
        )


def produce_force_constants(
    phonon: Phonopy,
    fc_calculator: Optional[str] = None,
    fc_calculator_options: Optional[str] = None,
    symmetrize_fc: bool = True,
    is_compact_fc: bool = True,
    log_level: int = 0,
):
    """Produce force constants."""
    try:
        phonon.produce_force_constants(
            calculate_full_force_constants=(not is_compact_fc),
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
        )
        if symmetrize_fc:
            phonon.symmetrize_force_constants(show_drift=(log_level > 0))
            if log_level:
                print("Force constants were symmetrized.")
    except ForcesetsNotFoundError:
        if log_level:
            print("Displacement-force datast was not found. ")


def check_nac_params(nac_params: dict, unitcell: PhonopyAtoms, pmat: np.ndarray):
    """Check number of Born effective charges."""
    borns = nac_params["born"]
    if len(borns) != np.rint(len(unitcell) * np.linalg.det(pmat)).astype(int):
        msg = "Number of Born effective charges is not consistent with the cell."
        raise ValueError(msg)


def _run_pypolymlp(
    phonon: Phonopy,
    mlp_params: Union[str, dict, PypolymlpParams],
    log_level: int = 0,
):
    """Run pypolymlp to compute forces."""
    if log_level:
        print("-" * 29 + " pypolymlp start " + "-" * 30)
        print("Pypolymlp is a generator of polynomial machine learning potentials.")
        print("Please cite the paper: A. Seko, J. Appl. Phys. 133, 011101 (2023).")
        print("Pypolymlp is developed at https://github.com/sekocha/pypolymlp.")
        if mlp_params:
            print("Parameters:")
            for k, v in asdict(parse_mlp_params(mlp_params)).items():
                if v is not None:
                    print(f"  {k}: {v}")

    mlp_loaded = False
    for mlp_filename in ["polymlp.yaml", "phonopy.pmlp"]:
        _mlp_filename_list = list(pathlib.Path().glob(f"{mlp_filename}*"))
        if _mlp_filename_list:
            _mlp_filename = _mlp_filename_list[0]
            if _mlp_filename.suffix not in [
                ".yaml",
                ".pmlp",
                ".xz",
                ".gz",
                ".bz2",
                "lzma",
            ]:
                continue
            if log_level:
                print(f'Load MLPs from "{_mlp_filename}".')
            phonon.load_mlp(_mlp_filename)
            mlp_loaded = True
            if log_level and mlp_filename == "phonopy.pmlp":
                print(f'Loading MLPs from "{_mlp_filename}" is obsolete.')
            break

    mlp_filename = "polymlp.yaml"
    if not mlp_loaded:
        if forces_in_dataset(phonon.mlp_dataset):
            if log_level:
                print("Developing MLPs by pypolymlp...", flush=True)
            phonon.develop_mlp(params=mlp_params)
            phonon.save_mlp()
            if log_level:
                print(f'MLPs were written into "{mlp_filename}"', flush=True)
        else:
            raise RuntimeError(f'"{mlp_filename}" is not found.')

    if log_level:
        print("-" * 30 + " pypolymlp end " + "-" * 31, flush=True)


def _prepare_dataset_by_pypolymlp(
    phonon: Phonopy,
    displacement_distance: Optional[float] = None,
    number_of_snapshots: Optional[Union[int, Literal["auto"]]] = None,
    random_seed: Optional[int] = None,
    log_level: int = 0,
):
    """Generate displacements and evaluate forces by pypolymlp."""
    if displacement_distance is None:
        _displacement_distance = 0.01
    else:
        _displacement_distance = displacement_distance

    if log_level:
        if number_of_snapshots:
            print("Generate random displacements")
            print(
                "  Twice of number of snapshots will be generated "
                "for plus-minus displacements."
            )
        else:
            print("Generate displacements")
        print(
            f"  Displacement distance: {_displacement_distance:.5f}".rstrip("0").rstrip(
                "."
            )
        )
    phonon.generate_displacements(
        distance=_displacement_distance,
        is_plusminus=True,
        number_of_snapshots=number_of_snapshots,
        random_seed=random_seed,
    )
    if log_level and number_of_snapshots == "auto":
        print(
            "  Number of generated supercells with random displacements: "
            f"{len(phonon.supercells_with_displacements)}",
        )

    if log_level:
        print(
            f"Evaluate forces in {len(phonon.displacements)} supercells by pypolymlp",
            flush=True,
        )

    if phonon.supercells_with_displacements is None:
        raise RuntimeError("Displacements are not set. Run generate_displacements.")

    phonon.evaluate_mlp()


def _read_force_constants_file(
    phonon: Phonopy, force_constants_filename, is_compact_fc=True, log_level=0
):
    dot_split = force_constants_filename.split(".")
    p2s_map = phonon.primitive.p2s_map
    if len(dot_split) > 1 and dot_split[-1] == "hdf5":
        _fc = read_force_constants_from_hdf5(
            filename=force_constants_filename,
            p2s_map=p2s_map,
            calculator=phonon.calculator,
        )
    else:
        _fc = parse_FORCE_CONSTANTS(filename=force_constants_filename, p2s_map=p2s_map)

    return _fc


def _read_crystal_structure(filename=None, interface_mode=None):
    try:
        return read_crystal_structure(filename=filename, interface_mode=interface_mode)
    except FileNotFoundError:
        raise
    except Exception as exc:
        msg = [
            "============================ phonopy.load ============================",
            "  Reading crystal structure file failed in phonopy.load.",
            "  Maybe phonopy.load(..., calculator='<calculator name>') expected?",
            "============================ phonopy.load ============================",
        ]
        raise RuntimeError("\n".join(msg)) from exc



"""Tests for band unfolding calculations."""

import os

import numpy as np

from phonopy import Phonopy
from phonopy.unfolding.core import Unfolding

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_Unfolding_NaCl(ph_nacl: Phonopy):
    """Test to reproduce proper band structure of primitive cell.

    Results are written to "bin-unfolding-test.dat".
    This data can be plotted by

    % plot_band.py bin-unfolding-test.dat

    Increase nd to get better plot.
    The test is done with nd=10.

    """
    # ph = _get_phonon(ph_nacl)
    ph = ph_nacl
    nd = 10
    qpoints = (
        np.array(
            [
                [
                    x,
                ]
                * 3
                for x in range(nd)
            ]
        )
        / float(nd)
        - 0.5
    )
    unfolding_supercell_matrix = [[-2, 2, 2], [2, -2, 2], [2, 2, -2]]
    mapping = np.arange(len(ph.supercell), dtype=int)
    unfolding = Unfolding(
        ph, unfolding_supercell_matrix, ph.supercell.scaled_positions, mapping, qpoints
    )
    unfolding.run()
    weights = _get_weights(unfolding, qpoints)
    # _write_weights(weights, "unfolding.dat")
    # filename_out = os.path.join(data_dir, "bin-unfolding-test.dat")
    _compare(weights, os.path.join(data_dir, "bin-unfolding.dat"), filename_out=None)


def test_Unfolding_SC(ph_nacl: Phonopy):
    """Test to reproduce unfoled band structure.

    Atomic positions are considered as the lattice ponts.

    Results are written to "bin-unfolding_to_atoms-test.dat".
    This data can be plotted by

    % plot_band.py bin-unfolding_to_atoms-test.dat

    Increase nd to get better plot.
    The test is done with nd=10.

    """
    # ph = _get_phonon(ph_nacl)
    ph = ph_nacl
    nd = 10
    qpoints = (
        np.array(
            [
                [
                    x,
                ]
                * 3
                for x in range(nd)
            ]
        )
        / float(nd)
        - 0.5
    )
    unfolding_supercell_matrix = np.diag([4, 4, 4])
    mapping = np.arange(len(ph.supercell), dtype=int)
    unfolding = Unfolding(
        ph, unfolding_supercell_matrix, ph.supercell.scaled_positions, mapping, qpoints
    )
    unfolding.run()
    weights = _get_weights(unfolding, qpoints)
    # _write_weights(weights, "unfolding_to_atoms.dat")
    # filename_out = os.path.join(data_dir, "bin-unfolding_to_atoms-test.dat")
    _compare(
        weights, os.path.join(data_dir, "bin-unfolding_to_atoms.dat"), filename_out=None
    )


def _compare(weights, filename, filename_out=None):
    bin_data = _binning(weights)
    if filename_out:
        _write_bin_data(bin_data, filename_out)
    with open(filename) as f:
        bin_data_in_file = np.loadtxt(f)
        np.testing.assert_allclose(bin_data, bin_data_in_file, atol=1e-2)


def _get_weights(unfolding, qpoints):
    weights = unfolding.unfolding_weights
    freqs = unfolding.frequencies

    out_vals = []
    for i, q in enumerate(qpoints):
        for f, w in zip(freqs[i], weights[i]):
            out_vals.append([q[0], q[1], q[2], f, w])

    return out_vals


def _write_weights(weights, filename):
    with open(filename, "w") as w:
        lines = ["%10.7f %10.7f %10.7f  %12.7f  %10.7f" % tuple(x) for x in weights]
        w.write("\n".join(lines))


def _write_bin_data(bin_data, filename):
    with open(filename, "w") as w:
        lines = ["%8.5f %8.5f %8.5f" % tuple(v) for v in bin_data]
        w.write("\n".join(lines))


def _binning(data):
    x = []
    y = []
    w = []
    for vals in data:
        if vals[4] > 1e-3:
            x.append(vals[0])
            y.append(vals[3])
            w.append(vals[4])
    x = np.around(x, decimals=5)
    y = np.around(y, decimals=5)
    w = np.array(w)

    points = {}
    for e_x, e_y, e_z in zip(x, y, w):
        if (e_x, e_y) in points:
            points[(e_x, e_y)] += e_z
        else:
            points[(e_x, e_y)] = e_z

    x = []
    y = []
    w = []
    for key in points:
        x.append(key[0])
        y.append(key[1])
        w.append(points[key])

    data = np.transpose([x, y, w])
    data = sorted(data, key=lambda data: data[1])
    data = sorted(data, key=lambda data: data[0])

    return np.array(data)


def _get_phonon(ph_in):
    ph = Phonopy(ph_in.supercell, supercell_matrix=[1, 1, 1])
    ph.force_constants = ph_in.force_constants
    born_elems = {
        s: ph_in.nac_params["born"][i] for i, s in enumerate(ph_in.primitive.symbols)
    }
    born = [born_elems[s] for s in ph_in.supercell.symbols]
    epsilon = ph_in.nac_params["dielectric"]
    factors = ph_in.nac_params["factor"]
    ph.nac_params = {"born": born, "factor": factors, "dielectric": epsilon}
    return ph



"""Tests for PhononMoment."""

from io import StringIO

import numpy as np

from phonopy import Phonopy
from phonopy.phonon.moment import PhononMoment

result_full_range = """
1.000000  1.000000  1.000000
 4.063222  4.236805  3.889623
17.935854 19.412820 16.458756
 1.000000  1.000000  1.000000
 3.530039  3.621065  3.451029
12.557720 13.205191 11.995720
"""


def test_moment(ph_nacl_nofcsym: Phonopy):
    """Test PhononMoment."""
    phonon = ph_nacl_nofcsym
    data = np.loadtxt(StringIO(result_full_range), dtype="double")
    moment = phonon.run_mesh(
        [13, 13, 13], with_eigenvectors=True, is_mesh_symmetry=False
    )
    num_atom = len(phonon.primitive)
    w = phonon.mesh.weights
    f = phonon.mesh.frequencies
    e = phonon.mesh.eigenvectors
    vals = np.zeros((6, num_atom + 1), dtype="double")

    moment = PhononMoment(f, w)
    for i in range(3):
        moment.run(order=i)
        vals[i, 0] = moment.moment
        assert np.abs(moment.moment - data[i, 0]) < 1e-5

    moment = PhononMoment(f, w, eigenvectors=e)
    for i in range(3):
        moment.run(order=i)
        moms = moment.moment
        vals[i, 1:] = moms
        assert (np.abs(moms - data[i, 1:]) < 1e-5).all()

    moment = PhononMoment(f, w)
    moment.set_frequency_range(freq_min=3, freq_max=4)
    for i in range(3):
        moment.run(order=i)
        vals[i + 3, 0] = moment.moment
        assert np.abs(moment.moment - data[i + 3, 0]) < 1e-5

    moment = PhononMoment(f, w, eigenvectors=e)
    moment.set_frequency_range(freq_min=3, freq_max=4)
    for i in range(3):
        moment.run(order=i)
        moms = moment.moment
        vals[i + 3, 1:] = moms
        assert (np.abs(moms - data[i + 3, 1:]) < 1e-5).all()

    # self._show(vals)


def _show(vals):
    for v in vals:
        print(("%9.6f " * len(v)) % tuple(v))



"""Phonopy loader."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import io
import os
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np

import phonopy.cui.load_helper as load_helper
from phonopy.api_phonopy import Phonopy
from phonopy.interface.calculator import get_calculator_physical_units
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_primitive_matrix
from phonopy.structure.dataset import forces_in_dataset


def load(
    phonopy_yaml: Optional[
        Union[str, bytes, os.PathLike, io.IOBase]
    ] = None,  # phonopy.yaml-like must be the first argument.
    supercell_matrix: Optional[Union[np.ndarray, Sequence]] = None,
    primitive_matrix: Optional[Union[np.ndarray, Sequence]] = None,
    is_nac: bool = True,
    calculator: Optional[str] = None,
    unitcell: Optional[PhonopyAtoms] = None,
    supercell: Optional[PhonopyAtoms] = None,
    nac_params: Optional[dict] = None,
    unitcell_filename: Optional[str] = None,
    supercell_filename: Optional[str] = None,
    born_filename: Optional[str] = None,
    force_sets_filename: Optional[str] = None,
    force_constants_filename: Optional[str] = None,
    fc_calculator: Optional[str] = None,
    fc_calculator_options: Optional[str] = None,
    factor: Optional[float] = None,
    produce_fc: bool = True,
    is_symmetry: bool = True,
    symmetrize_fc: bool = True,
    is_compact_fc: bool = True,
    use_pypolymlp: bool = False,
    mlp_params: Optional[dict] = None,
    store_dense_svecs: bool = True,
    use_SNF_supercell: bool = False,
    symprec: float = 1e-5,
    log_level: int = 0,
) -> Phonopy:
    """Create Phonopy instance from parameters and/or input files.

    "phonopy_yaml"-like file is parsed unless crystal structure information is
    given by unitcell_filename, supercell_filename, unitcell
    (PhonopyAtoms-like), or supercell (PhonopyAtoms-like). Even when
    "phonopy_yaml"-like file is parse, parameters except for crystal structure
    can be overwritten.

    Phonopy default files of 'FORCE_SETS' and 'BORN' are parsed when they are
    found in current directory and those data are not yet provided by other
    means.

    Crystal structure
    -----------------
    Means to provide crystal structure(s) and their priority:
        1. unitcell_filename (with supercell_matrix)
        2. supercell_filename
        3. unitcell (with supercell_matrix)
        4. supercell.
        5. phonopy_yaml

    Force sets or force constants
    -----------------------------
    Optional. Means to provide information to generate force constants and their
    priority:
        1. force_constants_filename
        2. force_sets_filename
        3. phonopy_yaml if force constants are found in phonoy_yaml.
        4. phonopy_yaml if forces are found in phonoy_yaml.dataset.
        5. 'FORCE_CONSTANTS' is searched in current directory.
        6. 'force_constants.hdf5' is searched in current directory.
        7. 'FORCE_SETS' is searched in current directory.
    When both of 3 and 4 are satisfied but not others, force constants and
    dataset are stored in Phonopy instance, but force constants are not produced
    from dataset.

    Parameters for non-analytical term correctiion (NAC)
    ----------------------------------------------------
    Optional. Means to provide NAC parameters and their priority:
        1. born_filename
        2. nac_params
        3. phonopy_yaml.nac_params if existed and is_nac=True.
        4. 'BORN' is searched in current directory when is_nac=True.

    Parameters
    ----------
    phonopy_yaml : str, bytes, os.PathLike, io.IOBase, optional
        Filename of "phonopy.yaml"-like file for str or bytes, otherwise file
        pointer-like. If this is given, the data in the file are parsed. Default
        is None.
    supercell_matrix : array_like, optional
        Supercell matrix multiplied to input cell basis vectors. shape=(3, ) or
        (3, 3), where the former is considered a diagonal matrix. Default is the
        unit matrix. dtype=int
    primitive_matrix : array_like or str, optional
        Primitive matrix multiplied to input cell basis vectors. Default is
        None, which is equivalent to 'auto'. For array_like, shape=(3, 3),
        dtype=float. When 'F', 'I', 'A', 'C', or 'R' is given instead of a 3x3
        matrix, the primitive matrix for the character found at
        https://spglib.github.io/spglib/definition.html is used.
    is_nac : bool, optional
        If True, look for 'BORN' file. If False, NAS is turned off. Default is
        True.
    calculator : str, optional.
        Calculator used for computing forces. This is used to switch the set of
        physical units. Default is None, which is equivalent to "vasp".
    unitcell : PhonopyAtoms, optional
        Input unit cell. Default is None.
    supercell : PhonopyAtoms, optional
        Input supercell. With given, default value of primitive_matrix is set to
        'auto' (can be overwitten). supercell_matrix is ignored. Default is
        None.
    nac_params : dict, optional
        Parameters required for non-analytical term correction. Default is None.
        {'born': Born effective charges
                 (array_like, shape=(primitive cell atoms, 3, 3), dtype=float),
         'dielectric': Dielectric constant matrix
                       (array_like, shape=(3, 3), dtype=float),
         'factor': unit conversion facotr (float)}
    unitcell_filename : str, optional
        Input unit cell filename. Default is None.
    supercell_filename : str, optional
        Input supercell filename. When this is specified, supercell_matrix is
        ignored. Default is None.
    born_filename : str, optional
        Filename corresponding to 'BORN', a file contains non-analytical term
        correction parameters.
    force_sets_filename : str, optional
        Filename of a file corresponding to 'FORCE_SETS', a file contains sets
        of forces and displacements. Default is None.
    force_constants_filename : str, optional
        Filename of a file corresponding to 'FORCE_CONSTANTS' or
        'force_constants.hdf5', a file contains force constants. Default is
        None.
    fc_calculator : str, optional
        Force constants calculator. Currently only 'alm'. Default is None.
    fc_calculator_options : str, optional
        Optional parameters that are passed to the external fc-calculator. This
        is given as one text string. How to parse this depends on the
        fc-calculator. For alm, each parameter is splitted by comma ',', and
        each set of key and value pair is written in 'key = value'.
    factor : float, optional
        Phonon frequency unit conversion factor. Unless specified, default unit
        conversion factor for each calculator is used.
    produce_fc : bool, optional
        Setting False, force constants are not calculated from dataset of
        displacements and forces even if the dataset exists. Default is True.
    is_symmetry : bool, optional
        Setting False, crystal symmetry except for lattice translation is not
        considered. Default is True.
    symmetrize_fc : bool, optional
        Setting False, force constants are not symmetrized when creating force
        constants from displacements and forces. Default is True.
    is_compact_fc : bool, optional
        Force constants are produced in the array whose shape is
            True: (primitive, supecell, 3, 3) False: (supercell, supecell, 3, 3)
        where 'supercell' and 'primitive' indicate number of atoms in these
        cells. Default is True.
    use_pypolymlp : bool, optional
        Use pypolymlp for generating force constants. Default is False.
    mlp_params : dict, optional
        A set of parameters used by machine learning potentials.
    store_dense_svecs : bool, optional
        Deprected. Dataset of shortest vectors between atoms in primitive
        cell and supercell is stored in the dense format when this is True.
        Default is True. In phonopy v3 or later version, False will not be
        supported.
    use_SNF_supercell : bool, optional
        Supercell is built by SNF algorithm when True. Default is False. SNF
        algorithm is faster than the original one, but the order of atoms in the
        supercell can be different from the original one. So the backward
        compatibility to the old data (e.g., force constants) may not be
        preserved.
    symprec : float, optional
        Tolerance used to find crystal symmetry. Default is 1e-5.
    log_level : int, optional
        Verbosity control. Default is 0.

    """
    if (
        supercell is not None
        or supercell_filename is not None
        or unitcell is not None
        or unitcell_filename is not None
    ):
        if primitive_matrix is None:
            _primitive_matrix = "auto"
        else:
            _primitive_matrix = primitive_matrix
        cell, smat, pmat = load_helper.get_cell_settings(
            supercell_matrix=supercell_matrix,
            primitive_matrix=_primitive_matrix,
            unitcell=unitcell,
            supercell=supercell,
            unitcell_filename=unitcell_filename,
            supercell_filename=supercell_filename,
            calculator=calculator,
            symprec=symprec,
            log_level=log_level,
        )
        _calculator = calculator
        _nac_params = nac_params
        _dataset = None
        _fc = None
    elif phonopy_yaml is not None:
        phpy_yaml = PhonopyYaml()
        phpy_yaml.read(phonopy_yaml)
        cell = phpy_yaml.unitcell
        smat = phpy_yaml.supercell_matrix
        if smat is None:
            smat = np.eye(3, dtype="intc", order="C")
        if primitive_matrix is not None:
            pmat = get_primitive_matrix(primitive_matrix, symprec=symprec)
        else:
            pmat = phpy_yaml.primitive_matrix
        if nac_params is not None:
            _nac_params = nac_params
        elif is_nac:
            _nac_params = phpy_yaml.nac_params
        else:
            _nac_params = None
        _dataset = phpy_yaml.dataset
        _fc = phpy_yaml.force_constants
        if calculator is None:
            _calculator = phpy_yaml.calculator
        else:
            _calculator = calculator
    else:
        msg = "Cell information could not found. Phonopy instance loading failed."
        raise RuntimeError(msg)

    if log_level and _calculator is not None:
        print('Set "%s" mode.' % _calculator)

    # units keywords: factor, nac_factor, distance_to_A
    units = get_calculator_physical_units(_calculator)
    if factor is None:
        _factor = units["factor"]
    else:
        _factor = factor
    phonon = Phonopy(
        cell,
        smat,
        primitive_matrix=pmat,
        factor=_factor,
        symprec=symprec,
        is_symmetry=is_symmetry,
        store_dense_svecs=store_dense_svecs,
        use_SNF_supercell=use_SNF_supercell,
        calculator=_calculator,
        log_level=log_level,
    )

    # NAC params
    if born_filename is not None or _nac_params is not None or is_nac:
        ret_nac_params = load_helper.get_nac_params(
            primitive=phonon.primitive,
            nac_params=_nac_params,
            born_filename=born_filename,
            is_nac=is_nac,
            nac_factor=units["nac_factor"],
            log_level=log_level,
        )
        if ret_nac_params is not None:
            phonon.nac_params = ret_nac_params

    dataset = load_helper.select_and_load_dataset(
        phonon,
        _dataset,
        force_sets_filename=force_sets_filename,
        log_level=log_level,
    )
    if dataset is not None:
        phonon.dataset = dataset

    fc = load_helper.select_and_extract_force_constants(
        phonon,
        fc=_fc,
        force_constants_filename=force_constants_filename,
        is_compact_fc=is_compact_fc,
        log_level=log_level,
    )
    if fc is not None:
        phonon.force_constants = fc

    if use_pypolymlp and dataset is not None:
        phonon.mlp_dataset = dataset
        phonon.dataset = None
        load_helper.prepare_pypolymlp_and_dataset(
            phonon,
            mlp_params=mlp_params,
            log_level=log_level,
        )
    if (
        phonon.force_constants is None
        and produce_fc
        and forces_in_dataset(phonon.dataset)
    ):
        load_helper.produce_force_constants(
            phonon,
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            symmetrize_fc=symmetrize_fc,
            is_compact_fc=is_compact_fc,
            log_level=log_level,
        )

    return phonon



"""A script to make FORCE_CONSTANTS from QE q2r.x output."""

import sys

from phonopy.interface.qe import PH_Q2R, read_pwscf

primcell_filename = sys.argv[1]
q2r_filename = sys.argv[2]
cell, _ = read_pwscf(primcell_filename)
q2r = PH_Q2R(q2r_filename)
q2r.run(cell)
q2r.write_force_constants()



"""PWmat calculator interface."""

# Copyright (C) 2014 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy as np

from phonopy.file_IO import iter_collect_forces
from phonopy.interface.vasp import check_forces, get_drift_forces
from phonopy.structure.atoms import PhonopyAtoms, atom_data


def parse_set_of_forces(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    hook = "force (eV/A)"
    is_parsed = True
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))
        pwmat_forces = iter_collect_forces(
            filename,
            num_atoms,
            hook,
            [1, 2, 3],
        )
        pwmat_forces = [[-x for x in sublist] for sublist in pwmat_forces]
        if check_forces(pwmat_forces, num_atoms, filename, verbose=verbose):
            drift_force = get_drift_forces(
                pwmat_forces, filename=filename, verbose=verbose
            )
            force_sets.append(np.array(pwmat_forces) - drift_force)
        else:
            is_parsed = False

    if is_parsed:
        return force_sets
    else:
        return []


def read_atom_config(filename):
    """Read structure information from atom.config."""
    with open(filename) as f:
        lines = f.read().splitlines()

    element_index = []
    lattice_vectors = []
    atom_position = []
    read_lattice = read_position = False

    for line in lines:
        line = line.strip()

        # lattice constant
        if line.lower().startswith("lattice"):
            read_lattice = True
        elif read_lattice:
            try:
                parts = [float(part) for part in line.split()[:3]]
                lattice_vectors.append(parts[0:3])
            except ValueError:
                read_lattice = False

        # atom position
        if line.lower().startswith("position"):
            read_position = True
        elif read_position:
            try:
                parts = [float(part) for part in line.split()]
                element_index.append(int(parts[0]))
                atom_position.append(parts[1:4])
            except ValueError:
                read_position = False

    cell_args = {
        "symbols": [atom_data[n][1] for n in element_index],
        "cell": lattice_vectors,
        "scaled_positions": atom_position,
    }

    return PhonopyAtoms(**cell_args)


def write_atom_config(filename, cell):
    """Write atom.config to file."""
    with open(filename, "w") as f:
        f.write(get_pwmat_structure(cell))


def get_pwmat_structure(cell):
    """Return PWmat structure in text."""
    lattice = cell.cell
    positions = cell.scaled_positions
    numbers = cell.numbers
    mag_mom = cell.magnetic_moments

    line = []
    line.append(f" {len(positions)}")
    line.append("Lattice vector (Angstrom)")
    for row in lattice:
        line.append(" ".join(f" {val:.16f}" for val in row))
    line.append("Position, move_x, move_y, move_z")
    for number, position in zip(numbers, positions):
        line.append(
            f" {number}  "
            + "   ".join(f"{val:.16f}" for val in position)
            + "  1   1   1"
        )
    if mag_mom is not None and len(mag_mom) == len(cell.numbers):
        if np.size(mag_mom[0]) == 1:
            line.append("magnetic")
            for number, mag in zip(numbers, mag_mom):
                line.append(f" {number}  {mag:.16f}")

        if np.size(mag_mom[0]) == 3:
            line.append("magnetic_xyz")
            for number, mag in zip(numbers, mag_mom):
                line.append(f" {number}  {mag[0]:.6f} {mag[1]:.6f} {mag[2]:.6f}")
    return "\n".join(line)


def write_supercells_with_displacements(
    supercell,
    cells_with_displacements,
    ids,
    pre_filename="supercell",
    width=3,
):
    """Write supercells with displacements to files."""
    write_atom_config(
        "%s.config" % pre_filename,
        supercell,
    )
    for i, cell in zip(ids, cells_with_displacements):
        filename = "{pre_filename}-{0:0{width}}.config".format(
            i, pre_filename=pre_filename, width=width
        )
        write_atom_config(filename, cell)



"""Tests for BrillouinZone class."""

import numpy as np
from spglib import get_stabilized_reciprocal_mesh, relocate_BZ_grid_address

from phonopy.structure.brillouin_zone import BrillouinZone


def test_FCC():
    """Test BrillouinZone with FCC lattice."""
    direct_lat = [[0, 2.73, 2.73], [2.73, 0, 2.73], [2.73, 2.73, 0]]
    is_shift = [0, 0, 0]
    _testBrillouinZone(direct_lat, [4, 4, 4], is_shift)
    _testBrillouinZone(direct_lat, [5, 5, 5], is_shift)


def test_Hexagonal():
    """Test BrillouinZone with Hexagonal lattice."""
    direct_lat = [
        [3.0751691007292523, 0, 0],
        [-1.5375845503646262, 2.6631745621644800, 0],
        [0, 0, 3.5270080068586522],
    ]
    is_shift = [0, 0, 0]
    _testBrillouinZone(direct_lat, [4, 4, 4], is_shift)
    _testBrillouinZone(direct_lat, [5, 5, 5], is_shift)


def _testBrillouinZone(direct_lat, mesh, is_shift):
    _, grid_address = get_stabilized_reciprocal_mesh(
        mesh,
        rotations=[
            np.eye(3, dtype="intc"),
        ],
        is_shift=is_shift,
    )
    rec_lat = np.linalg.inv(direct_lat)
    bz_grid_address, bz_map = relocate_BZ_grid_address(
        grid_address, mesh, rec_lat, is_shift=is_shift
    )

    qpoints = (grid_address + np.array(is_shift) / 2.0) / mesh
    bz = BrillouinZone(rec_lat)
    bz.run(qpoints)
    sv_all = bz.shortest_qpoints  # including BZ boundary duplicates
    sv = [v[0] for v in sv_all]
    bz_qpoints = (bz_grid_address + np.array(is_shift) / 2.0) / mesh
    d2_this = (np.dot(sv, rec_lat.T) ** 2).sum(axis=1)
    d2_spglib = (np.dot(bz_qpoints[: np.prod(mesh)], rec_lat.T) ** 2).sum(axis=1)
    diff = d2_this - d2_spglib
    diff -= np.rint(diff)

    # Following both of two tests are necessary.
    # Check equivalence of vectors by lattice translation
    np.testing.assert_allclose(diff, 0, atol=1e-8)
    # Check being in same (hopefull first) Brillouin zone by their lengths
    np.testing.assert_allclose(d2_this, d2_spglib, atol=1e-8)



"""Tests for PhonopyAtoms."""

from collections.abc import Callable
from io import StringIO

import numpy as np
import pytest
import yaml

from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms, parse_cell_dict

symbols_SiO2 = ["Si"] * 2 + ["O"] * 4
symbols_AcO2 = ["Ac"] * 2 + ["O"] * 4
_lattice = [[4.65, 0, 0], [0, 4.75, 0], [0, 0, 3.25]]
_points = [
    [0.0, 0.0, 0.0],
    [0.5, 0.5, 0.5],
    [0.3, 0.3, 0.0],
    [0.7, 0.7, 0.0],
    [0.2, 0.8, 0.5],
    [0.8, 0.2, 0.5],
]
cell_SiO2 = PhonopyAtoms(cell=_lattice, scaled_positions=_points, symbols=symbols_SiO2)
cell_AcO2 = PhonopyAtoms(cell=_lattice, scaled_positions=_points, symbols=symbols_AcO2)


def test_SiO2():
    """Test of attributes by SiO2."""
    _test_cell(cell_SiO2, _lattice, _points, symbols_SiO2)


def test_SiO2_copy(helper_methods: Callable):
    """Test of PhonopyAtoms.copy() by SiO2."""
    helper_methods.compare_cells(cell_SiO2, cell_SiO2.copy())
    helper_methods.compare_cells_with_order(cell_SiO2, cell_SiO2.copy())


def test_AcO2(helper_methods: Callable):
    """Test of attributes by AcO2."""
    _test_cell(cell_AcO2, _lattice, _points, symbols_AcO2)
    helper_methods.compare_cells(cell_AcO2, cell_AcO2.copy())
    helper_methods.compare_cells_with_order(cell_AcO2, cell_AcO2.copy())


def test_AcO2_copy():
    """Test of PhonopyAtoms.copy() by AcO2."""


def _test_cell(cell: PhonopyAtoms, lattice, points, symbols):
    np.testing.assert_allclose(cell.cell, lattice, atol=1e-8)
    for s1, s2 in zip(cell.symbols, symbols):
        assert s1 == s2
    diff = cell.scaled_positions - points
    diff -= np.rint(diff)
    dist = np.linalg.norm(np.dot(diff, cell.cell), axis=1)
    np.testing.assert_allclose(dist, np.zeros(len(dist)), atol=1e-8)


def test_phonopy_atoms_SiO2():
    """Test of PhonopyAtoms __str__ by SiO2."""
    _test_phonopy_atoms(cell_SiO2)


def test_phonopy_atoms_AcO2():
    """Test of PhonopyAtoms __str__ by AcO2."""
    _test_phonopy_atoms(cell_AcO2)


@pytest.mark.parametrize(
    "is_ncl,is_flat", [(False, False), (False, True), (True, False), (True, True)]
)
def test_Cr_magnetic_moments(convcell_cr: PhonopyAtoms, is_ncl: bool, is_flat: bool):
    """Test by Cr with magnetic moments."""
    print(len(convcell_cr))
    if is_ncl:
        if is_flat:
            convcell_cr.magnetic_moments = [0, 0, 1, 0, 0, -1]
        else:
            convcell_cr.magnetic_moments = [[0, 0, 1], [0, 0, -1]]
        np.testing.assert_allclose(
            convcell_cr.magnetic_moments, [[0, 0, 1], [0, 0, -1]]
        )
    else:
        if is_flat:
            convcell_cr.magnetic_moments = [1, -1]
        else:
            convcell_cr.magnetic_moments = [[1], [-1]]
        np.testing.assert_allclose(convcell_cr.magnetic_moments, [1, -1])
    _test_phonopy_atoms(convcell_cr)
    convcell_cr.magnetic_moments = None


@pytest.mark.parametrize("is_ncl", [False, True])
def test_Cr_copy_magnetic_moments(
    convcell_cr: PhonopyAtoms, is_ncl: bool, helper_methods
):
    """Test by Cr with magnetic moments."""
    if is_ncl:
        convcell_cr.magnetic_moments = [[0, 0, 1], [0, 0, -1]]
    else:
        convcell_cr.magnetic_moments = [1, -1]
    helper_methods.compare_cells(convcell_cr, convcell_cr.copy())
    helper_methods.compare_cells_with_order(convcell_cr, convcell_cr.copy())
    convcell_cr.magnetic_moments = None


def test_parse_cell_dict(helper_methods: Callable):
    """Test parse_cell_dict."""
    cell = cell_SiO2
    points = []
    for coord, mass, symbol in zip(cell.scaled_positions, cell.masses, cell.symbols):
        points.append({"symbol": symbol, "coordinates": coord, "mass": mass})
    cell_dict = {"lattice": cell_SiO2.cell, "points": points}
    _cell = parse_cell_dict(cell_dict)
    helper_methods.compare_cells_with_order(cell, _cell)


def test_PhonopyAtoms_with_Xn_symbol(ph_nacl: Phonopy):
    """Test of PhonopyAtoms with Xn symbol."""
    symbols = ph_nacl.unitcell.symbols
    symbols[-1] = "Cl1"
    masses = ph_nacl.unitcell.masses
    masses[-1] = 70.0
    numbers = ph_nacl.unitcell.numbers
    numbers[-1] = numbers[-1] + PhonopyAtoms._MOD_DIVISOR

    with pytest.raises(RuntimeError) as e:
        _ = PhonopyAtoms(
            cell=ph_nacl.unitcell.cell,
            scaled_positions=ph_nacl.unitcell.scaled_positions,
            symbols=symbols,
        )
    assert str(e.value) == "Masses have to be specified when special symbols are used."

    with pytest.raises(RuntimeError) as e:
        _ = PhonopyAtoms(
            cell=ph_nacl.unitcell.cell,
            scaled_positions=ph_nacl.unitcell.scaled_positions,
            numbers=numbers,
        )
    assert str(e.value) == "Atomic numbers cannot be larger than 118."

    symbols[-1] = "Cl_1"
    with pytest.raises(RuntimeError) as e:
        _ = PhonopyAtoms(
            cell=ph_nacl.unitcell.cell,
            scaled_positions=ph_nacl.unitcell.scaled_positions,
            symbols=symbols,
        )
    assert str(e.value) == "Invalid symbol: Cl_1."

    symbols[-1] = "Cl_0"
    with pytest.raises(RuntimeError) as e:
        _ = PhonopyAtoms(
            cell=ph_nacl.unitcell.cell,
            scaled_positions=ph_nacl.unitcell.scaled_positions,
            symbols=symbols,
        )
    assert str(e.value) == "Invalid symbol: Cl_0."


def _test_phonopy_atoms(cell: PhonopyAtoms):
    with StringIO(str(cell.copy())) as f:
        data = yaml.safe_load(f)
        np.testing.assert_allclose(cell.cell, data["lattice"], atol=1e-8)
        positions = []
        magmoms = []
        for atom, symbol in zip(data["points"], cell.symbols):
            positions.append(atom["coordinates"])
            if "magnetic_moment" in atom:
                magmoms.append(atom["magnetic_moment"])
            assert atom["symbol"] == symbol

        diff = cell.scaled_positions - positions
        diff -= np.rint(diff)
        dist = np.linalg.norm(np.dot(diff, cell.cell), axis=1)
        np.testing.assert_allclose(dist, np.zeros(len(dist)), atol=1e-8)

        if magmoms:
            np.testing.assert_allclose(cell.magnetic_moments, magmoms, atol=1e-8)


def test_formula():
    """Test of PhonopyAtoms formula property."""
    # Test SiO2
    assert cell_SiO2.formula == "O4Si2"

    # Test AcO2
    assert cell_AcO2.formula == "Ac2O4"

    # Test with indexed symbols
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
        symbols=["Fe", "Fe1"],
        masses=[55.845, 55.845],  # Add masses for Fe
    )
    assert cell.formula == "Fe2"

    # Test single atom
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0]],
        symbols=["H"],
    )
    assert cell.formula == "H"

    # Test edge cases
    # Empty cell
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[],
        symbols=[],
    )
    assert cell.formula == ""

    # Multiple digit indices
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0]],
        symbols=["Fe11", "Fe2"],
        masses=[55.845, 55.845],
    )
    assert cell.formula == "Fe2"


def test_formula_complex():
    """Test formula property with more complex structures."""
    # Test structure with multiple elements
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0], [0, 0, 0.5]],
        symbols=["Fe", "O", "O", "Ti"],
    )
    assert cell.formula == "FeO2Ti"

    # Test structure with indexed symbols
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0]],
        symbols=["Fe1", "Fe2", "O"],
        masses=[55.845, 55.845, 15.999],  # Add masses for Fe and O
    )
    assert cell.formula == "Fe2O"


def test_reduced_formula():
    """Test reduced_formula property."""
    # Test reduction of larger numbers
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0], [0.5, 0.5, 0]],
        symbols=["Fe", "Fe", "O", "O"],
    )
    assert cell.formula == "Fe2O2"
    assert cell.reduced_formula == "FeO"

    # Test no reduction needed
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0]],
        symbols=["Fe", "O", "O"],
    )
    assert cell.formula == "FeO2"
    assert cell.reduced_formula == "FeO2"

    # Test with indexed symbols
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0], [0.5, 0.5, 0]],
        symbols=["Fe1", "Fe2", "O1", "O2"],
        masses=[55.845, 55.845, 15.999, 15.999],  # Add masses for Fe and O
    )
    assert cell.formula == "Fe2O2"
    assert cell.reduced_formula == "FeO"

    # Test single atom
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0]],
        symbols=["H"],
    )
    assert cell.formula == "H"
    assert cell.reduced_formula == "H"

    # Test edge cases
    # Empty cell
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[],
        symbols=[],
    )
    assert cell.reduced_formula == ""

    # Large numbers that reduce
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0]] * 12,
        symbols=["Fe"] * 6 + ["O"] * 6,
    )
    assert cell.formula == "Fe6O6"
    assert cell.reduced_formula == "FeO"

    # Prime numbers that don't reduce
    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0]] * 5,
        symbols=["Fe"] * 2 + ["O"] * 3,
    )
    assert cell.formula == "Fe2O3"
    assert cell.reduced_formula == "Fe2O3"


@pytest.mark.parametrize(
    "symbols,expected_formula,expected_normalized",
    [
        (["Fe", "Fe", "O", "O"], "Fe2O2", "Fe0.5O0.5"),
        (["Fe", "O", "O"], "FeO2", "Fe0.333O0.667"),
        (["Fe", "Fe", "O", "O", "O"], "Fe2O3", "Fe0.4O0.6"),
        (["H"], "H", "H1.0"),
        ([], "", ""),
        (["Fe"] * 6 + ["O"] * 6, "Fe6O6", "Fe0.5O0.5"),
        # Test with indexed symbols
        (["Fe1", "Fe2", "O"], "Fe2O", "Fe0.667O0.333"),
    ],
)
def test_formulas(symbols, expected_formula, expected_normalized):
    """Test all formula properties."""
    # Create cell with appropriate masses for indexed symbols
    masses = None
    if any(s.rstrip("0123456789") != s for s in symbols):
        masses = []
        for s in symbols:
            base = s.rstrip("0123456789")
            if base == "Fe":
                masses.append(55.845)
            elif base == "O":
                masses.append(15.999)
            elif base == "H":
                masses.append(1.008)

    cell = PhonopyAtoms(
        cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        scaled_positions=[[0, 0, 0]] * len(symbols),
        symbols=symbols,
        masses=masses,
    )

    assert cell.formula == expected_formula
    assert cell.normalized_formula == expected_normalized



"""Tools to manage displacement dataset."""

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from typing import Optional

import numpy as np


def get_displacements_and_forces(
    disp_dataset: dict,
) -> tuple[np.ndarray, Optional[np.ndarray]]:
    """Return displacements and forces of all atoms from displacement dataset.

    This is used to extract displacements and forces from displacement dataset.
    This method is considered more-or-less as a converter when the input is in
    type-1.

    Parameters
    ----------
    disp_dataset : dict
        Displacement dataset either in type-1 or type-2.

    Returns
    -------
    displacements : ndarray
        Displacements of all atoms in all supercells.
        shape=(snapshots, supercell atoms, 3), dtype='double', order='C'
    forces : ndarray or None
        Forces of all atoms in all supercells.
        shape=(snapshots, supercell atoms, 3), dtype='double', order='C'
        None is returned when forces don't exist.

    """
    if "first_atoms" in disp_dataset:
        natom = disp_dataset["natom"]
        disps = np.zeros(
            (len(disp_dataset["first_atoms"]), natom, 3), dtype="double", order="C"
        )
        forces = None
        for i, disp1 in enumerate(disp_dataset["first_atoms"]):
            disps[i, disp1["number"]] = disp1["displacement"]
            if "forces" in disp1:
                if forces is None:
                    forces = np.zeros_like(disps)
                forces[i] = disp1["forces"]
        return disps, forces
    elif "displacements" in disp_dataset:
        if "forces" in disp_dataset:
            forces = disp_dataset["forces"]
        else:
            forces = None
        return disp_dataset["displacements"], forces


def forces_in_dataset(dataset):
    """Check if forces in displacement dataset."""
    if dataset is None:
        return False

    if not isinstance(dataset, dict):
        raise RuntimeError("dataset is wrongly made.")

    if "first_atoms" in dataset:  # type-1
        for d in dataset["first_atoms"]:
            if "forces" not in d:
                return False
        return True

    if "forces" in dataset:  # type-2
        return True

    return False



"""Linear tetrahedron method on regular grid."""

# copyright (C) 2013 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.tetrahedron_method import TetrahedronMethod


class TetrahedronMesh:
    """Class to perform linear tetrahedron method on regular grid."""

    def __init__(
        self,
        cell: PhonopyAtoms,
        frequencies,  # only at ir-grid-points
        mesh,
        grid_address,
        grid_mapping_table,
        ir_grid_points,
        grid_order=None,
        lang="C",
    ):
        """Linear tetrahedron method on uniform mesh for phonons.

        Parameters
        ----------
        cell : PhonopyAtoms
            Primitive cell used to calculate frequencies
        frequencies: ndarray
            Phonon frequences on ir-grid points
            shape=(num_ir_grid_points, num_band)
            dtype='double'
        mesh : ndarray or list of int
            Mesh numbers for grids
            shape=(3,)
            dtype='int64'
        grid_address : ndarray
            Addresses of all grid points given by GridPoints class.
            shape=(prod(mesh), 3)
            dtype='int64'
        grid_mapping_table : ndarray
            Mapping of grid points to irreducible grid points given by
            GridPoints class.
            shape=(prod(mesh),)
            dtype='int64'
        ir_grid_points : ndarray
            Irreducible gird points given by GridPoints class.
            shape=(len(np.unique(grid_mapping_table)),)
            dtype='int64'
        grid_order : list of int, optional
            This controls how grid addresses are stored either C style or
            Fortran style.
        lang : str, 'C' or else, optional
            With 'C', C implementation is used. Otherwise Python implementation
            runs.

        """
        self._cell = cell
        self._frequencies = frequencies
        self._mesh = np.array(mesh, dtype="int64")
        self._grid_address = grid_address
        self._grid_mapping_table = grid_mapping_table
        self._lang = lang
        if lang == "C":
            self._grid_order = None
        else:
            if grid_order is None:
                self._grid_order = [1, mesh[0], mesh[0] * mesh[1]]
            else:
                self._grid_order = grid_order
        self._ir_grid_points = ir_grid_points

        self._gp_ir_index = None

        self._tm = None
        self._tetrahedra_frequencies = None
        self._integration_weights = None
        self._relative_grid_address = None

        self._frequency_points = None
        self._value = None

        self._grid_point_count = 0

        self._prepare()

    def __iter__(self):
        """Define iterator over grid points."""
        return self

    def __next__(self):
        """Peform linear tetrahedron method at a grid point."""
        if self._grid_point_count == len(self._ir_grid_points):
            raise StopIteration
        else:
            gp = self._ir_grid_points[self._grid_point_count]
            self._set_tetrahedra_frequencies(gp)
            for ib, frequencies in enumerate(self._tetrahedra_frequencies):
                self._tm.set_tetrahedra_omegas(frequencies)
                self._tm.run(self._frequency_points, value=self._value)
                iw = self._tm.get_integration_weight()
                self._integration_weights[:, ib] = iw
            self._integration_weights /= np.prod(self._mesh)
            self._grid_point_count += 1
            return self._integration_weights

    def get_integration_weights(self):
        """Return integration weights."""
        return self._integration_weights

    def get_frequency_points(self):
        """Return frequency points."""
        return self._frequency_points

    def set(self, value="I", division_number=201, frequency_points=None, lang="C"):
        """Prepare environment to peform linear tetrahedron method."""
        self._grid_point_count = 0
        self._value = value
        if frequency_points is None:
            max_frequency = np.amax(self._frequencies)
            min_frequency = np.amin(self._frequencies)
            self._frequency_points = np.linspace(
                min_frequency, max_frequency, division_number, dtype="double"
            )
        else:
            self._frequency_points = np.array(frequency_points, dtype="double")

        num_band = self._frequencies.shape[1]
        num_freqs = len(self._frequency_points)
        self._integration_weights = np.zeros((num_freqs, num_band), dtype="double")
        reciprocal_lattice = np.linalg.inv(self._cell.cell)
        self._tm = TetrahedronMethod(reciprocal_lattice, mesh=self._mesh, lang=lang)
        self._relative_grid_address = self._tm.tetrahedra

    def _prepare(self):
        ir_gp_indices = {}
        for i, gp in enumerate(self._ir_grid_points):
            ir_gp_indices[gp] = i

        self._gp_ir_index = np.zeros_like(self._grid_mapping_table)
        for i, gp in enumerate(self._grid_mapping_table):
            self._gp_ir_index[i] = ir_gp_indices[gp]

    def _set_tetrahedra_frequencies(self, gp):
        self._tetrahedra_frequencies = get_tetrahedra_frequencies(
            gp,
            self._mesh,
            self._grid_address,
            self._relative_grid_address,
            self._gp_ir_index,
            self._frequencies,
            grid_order=self._grid_order,
            lang=self._lang,
        )


def get_tetrahedra_frequencies(
    gp,
    mesh,
    grid_address,
    relative_grid_address,
    gp_ir_index,
    frequencies,
    grid_order=None,
    lang="C",
):
    """Return frequencies on the relative_grid_addresses.

    Note
    ----
    This implementation is based on GR-grid.

    Parameters
    ----------
    gp : float
        Grid index
    mesh : ndarray
        Mesh numbers. shape=(3, ), dtype='int64'
    grid_address : ndarray
        Grid address in integers. shape=(prod(mesh), 3), dtype='int64', order='C'
    relative_grid_addresses : ndarray
        Relative grid addresses from the centre (i.e., gp) shape=(24, 4, 3),
        dtype='int64', order='C'
    gp_ir_index : ndarray
        Mapping table from grid index in GR-grid to index corresponding to first
        dimension of frequencies. The ir-grid index is
        range(len(ir-grid-points)). shape=(prod(mesh), ), dtype='int64'
    frequencies : ndarray
        Phonon frequences on ir-grid points. shape=(ir-grid-points, num_band)
        dtype='double'
    grid_order : list of int, optional
        This controls how grid addresses are stored either C style or Fortran
        style. This is only valid when lang != 'C'.
    lang : str, 'C' or else, optional
        With 'C', C implementation is used. Otherwise Python implementation
        runs.

    Returns
    -------
    ndarray
        Frequencies at tetheredra tertices. shape=(num_bands, 24, 4),
        dtype='double', order='C'

    """
    if lang == "C":
        try:
            import phonopy._phonopy as phonoc  # noqa F401

            return _get_tetrahedra_frequencies_C(
                gp, mesh, grid_address, relative_grid_address, gp_ir_index, frequencies
            )
        except ImportError:
            return _get_tetrahedra_frequencies_Py(
                gp,
                mesh,
                grid_address,
                relative_grid_address,
                gp_ir_index,
                frequencies,
                grid_order,
            )
    else:
        return _get_tetrahedra_frequencies_Py(
            gp,
            mesh,
            grid_address,
            relative_grid_address,
            gp_ir_index,
            frequencies,
            grid_order,
        )


def _get_tetrahedra_frequencies_C(
    gp, mesh, grid_address, relative_grid_address, gp_ir_index, frequencies
):
    import phonopy._phonopy as phonoc

    t_frequencies = np.zeros((1, frequencies.shape[1], 24, 4), dtype="double")
    phonoc.tetrahedra_frequencies(
        t_frequencies,
        np.array([gp], dtype="int64"),
        mesh,
        grid_address,
        gp_ir_index,
        relative_grid_address,
        frequencies,
    )
    return np.array(t_frequencies[0], dtype="double", order="C")


def _get_tetrahedra_frequencies_Py(
    gp, mesh, grid_address, relative_grid_address, gp_ir_index, frequencies, grid_order
):
    t_frequencies = np.zeros((frequencies.shape[1], 24, 4), dtype="double")
    for i, t in enumerate(relative_grid_address):
        address = t + grid_address[gp]
        neighbors = np.dot(address % mesh, grid_order)
        t_frequencies[:, i, :] = frequencies[gp_ir_index[neighbors]].T
    return t_frequencies



# Copyright (C) 2025 Eugene Roginskii
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import argparse
import sys

import numpy as np

from phonopy.api_phonopy import Phonopy
from phonopy.interface.cp2k import read_cp2k
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.structure.symmetry import Symmetry


def genxyz(fn, atnum, basis, species, cart, desc=""):
    """Generate an .xyz file with the given structure."""
    print("Generating file %s" % fn)
    try:
        fh = open(fn, "w")
    except IOError:
        print("ERROR Couldn't open output file %s for writing" % fn)
        return -1
    fh.write("%d\n" % len(species))
    fh.write("%s\n" % "Force calculations. %s" % desc)
    for i in range(len(species)):
        fh.write(
            "%s  %s\n" % (species[i], "".join("    % 15.10f" % c for c in cart[i]))
        )


parser = argparse.ArgumentParser(
    description="""Input files for phonon
                                 calculations in .xyz format generator"""
)

parser.add_argument(
    "-i",
    "--input",
    action="store",
    type=str,
    dest="str_fn",
    help="Input filename with structure",
)
parser.add_argument(
    "-o",
    "--output",
    action="store",
    type=str,
    dest="out_fn",
    default="supercell",
    help="Output filename prefix",
)
parser.add_argument(
    "-d",
    "--dim",
    action="store",
    type=str,
    dest="dim",
    default="1 1 1",
    help="Supercell dimentions",
)
parser.add_argument(
    "-a",
    "--amplitude",
    action="store",
    type=float,
    dest="ampl",
    default=0.01,
    help="Displacement amplitude",
)
parser.add_argument(
    "--born",
    dest="born",
    action="store_true",
    default=False,
    help="Generate supercells for BORN charges calculations",
)
parser.add_argument(
    "--nosym", dest="nosym", action="store_true", default=False, help="Ignore symmetry"
)

args = parser.parse_args()

if args.str_fn is None:
    print("Error. No input filename was given.")
    sys.exit(1)
if len(args.dim.split()) < 3:
    print("Size of dim should be 3 or 9")

if len(args.dim.split()) > 0:
    if len(args.dim.split()) > 3:
        dim = np.array([float(d) for d in args.dim.split()]).reshape(3, 3)
    else:
        dim = np.zeros(9).reshape(3, 3)
        for i in range(3):
            dim[i][i] = float(args.dim.split()[i])
else:
    dim = np.eye(3)

inph = read_cp2k(args.str_fn)
species = inph[0].symbols
numbers = inph[0].numbers
masses = inph[0].masses
natom = len(inph[0])
basis = inph[0].cell
cart = inph[0].positions
cvol = inph[0].volume
ph = Phonopy(inph[0], supercell_matrix=dim)
scideal = ph.supercell
displs = []

scMap = ph.symmetry.get_map_atoms()
s2uMap = ph._primitive.p2p_map

if args.nosym:
    print("Displacements with ignore symmetry option")
    # Iterate over all unique atoms in supercell mapped to unitcell ones
    for atom in s2uMap:
        for i in range(3):
            displ = [0.0, 0.0, 0.0]
            displ[i] = args.ampl
            displs.append({"number": atom, "displacement": displ})
            displ = [0.0, 0.0, 0.0]
            displ[i] = -args.ampl
            displs.append({"number": atom, "displacement": displ})
            ph.dataset = {"natom": len(scideal), "first_atoms": displs}
elif args.born:
    for atom in Symmetry(scideal).get_independent_atoms():
        for i in range(3):
            displ = [0.0, 0.0, 0.0]
            displ[i] = args.ampl
            displs.append({"number": atom, "displacement": displ})
            displ = [0.0, 0.0, 0.0]
            displ[i] = -args.ampl
            displs.append({"number": atom, "displacement": displ})
            ph.dataset = {"natom": len(scideal), "first_atoms": displs}
else:
    ph.generate_displacements(distance=args.ampl, is_plusminus="auto", is_diagonal=True)

scs = ph.supercells_with_displacements

scideal = ph.supercell

fn = "".join("%s-%04d.xyz" % (args.out_fn, 0))
genxyz(fn, len(scideal), scideal.cell, scideal.symbols, scideal.positions)

cellstr = "A " + "%14.8f " * 3 % tuple(scideal.cell[0])
cellstr += "B " + "%14.8f " * 3 % tuple(scideal.cell[1])
cellstr += "C " + "%14.8f " * 3 % tuple(scideal.cell[2])

for i in range(len(scs)):
    fn = "".join("%s-%04d.xyz" % (args.out_fn, i + 1))
    genxyz(
        fn,
        len(scs[i]),
        scs[i].cell,
        scs[i].symbols,
        scs[i].positions,
        desc="Lattice parameters: %s" % cellstr,
    )

print("Lattice parameters to put in pattern input file:")
for i, abc in enumerate(scideal.cell):
    print("%s " % "ABC"[i], "% 14.8f " * 3 % tuple(abc))

fn = "".join("%s-%04d.xyz" % (args.out_fn, 0))
settings = {
    "force_sets": False,
    "displacements": True,
    "force_constants": False,
    "born_effective_charge": True,
    "dielectric_constant": True,
}

conf = {"DIM": args.dim}
phyaml = PhonopyYaml(configuration=conf, calculator="cp2k", settings=settings)
phyaml.set_phonon_info(ph)


with open("phonopy_disp.yaml", "w") as fd:
    fd.write(phyaml.__str__())



"""Tests for wien2k interface."""

import os

import numpy as np

from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.interface.wien2k import parse_wien2k_struct

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_parse_wien2k_struct():
    """Test structure parsing."""
    filename_BaGa2 = os.path.join(data_dir, "BaGa2.struct")
    cell, _, _, _ = parse_wien2k_struct(filename_BaGa2)
    filename = os.path.join(data_dir, "BaGa2-wien2k.yaml")
    cell_ref = read_cell_yaml(filename)
    assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r



"""ABACUS calculator interface."""

# Copyright (C) 2022 Yuyang Ji
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
import sys
from collections import Counter

import numpy as np

from phonopy.interface.vasp import check_forces, get_drift_forces
from phonopy.structure.atoms import PhonopyAtoms, atom_data, symbol_map

_re_float = r"[-+]?\d+\.*\d*(?:[Ee][-+]\d+)?"

#
# read ABACUS STRU
#


def read_abacus(filename):
    """Read structure information, distance in unit au (bohr)."""
    fd = open(filename, "r")
    contents = fd.read()
    title_str = (
        r"(?:LATTICE_CONSTANT|NUMERICAL_ORBITAL|ABFS_ORBITAL|"
        + r"LATTICE_VECTORS|LATTICE_PARAMETERS|ATOMIC_POSITIONS)"
    )

    # remove comments and empty lines
    contents = re.compile(r"#.*|//.*").sub("", contents)
    contents = re.compile(r"\n{2,}").sub("\n", contents)

    # specie, mass, pps
    specie_pattern = re.compile(rf"ATOMIC_SPECIES\s*\n([\s\S]+?)\s*\n{title_str}")
    specie_lines = np.array(
        [line.split() for line in specie_pattern.search(contents).group(1).split("\n")]
    )
    symbols = specie_lines[:, 0]
    ntype = len(symbols)
    try:
        atom_potential = dict(zip(symbols, specie_lines[:, 2].tolist()))
    except IndexError:
        atom_potential = None

    # basis
    aim_title = "NUMERICAL_ORBITAL"
    aim_title_sub = title_str.replace("|" + aim_title, "")
    orb_pattern = re.compile(rf"{aim_title}\s*\n([\s\S]+?)\s*\n{aim_title_sub}")
    orb_lines = orb_pattern.search(contents)
    if orb_lines:
        atom_basis = dict(zip(symbols, orb_lines.group(1).split("\n")))
    else:
        atom_basis = None

    # ABFs basis
    aim_title = "ABFS_ORBITAL"
    aim_title_sub = title_str.replace("|" + aim_title, "")
    abf_pattern = re.compile(rf"{aim_title}\s*\n([\s\S]+?)\s*\n{aim_title_sub}")
    abf_lines = abf_pattern.search(contents)
    if abf_lines:
        atom_offsite_basis = dict(zip(symbols, abf_lines.group(1).split("\n")))
    else:
        atom_offsite_basis = None

    # lattice constant
    aim_title = "LATTICE_CONSTANT"
    aim_title_sub = title_str.replace("|" + aim_title, "")
    a0_pattern = re.compile(rf"{aim_title}\s*\n([\s\S]+?)\s*\n{aim_title_sub}")
    a0_lines = a0_pattern.search(contents)
    atom_lattice_scale = float(a0_lines.group(1))

    # lattice vector
    aim_title = "LATTICE_VECTORS"
    aim_title_sub = title_str.replace("|" + aim_title, "")
    vec_pattern = re.compile(rf"{aim_title}\s*\n([\s\S]+?)\s*\n{aim_title_sub}")
    vec_lines = vec_pattern.search(contents)
    if vec_lines:
        atom_lattice = np.array(
            [line.split() for line in vec_pattern.search(contents).group(1).split("\n")]
        ).astype(float)
    atom_lattice = atom_lattice * atom_lattice_scale

    aim_title = "ATOMIC_POSITIONS"
    type_pattern = re.compile(rf"{aim_title}\s*\n(\w+)\s*\n")
    # type of coordinates
    atom_pos_type = type_pattern.search(contents).group(1)
    assert atom_pos_type in [
        "Direct",
        "Cartesian",
    ], "Only two type of atomic coordinates are supported: 'Direct' or 'Cartesian'."

    block_pattern = re.compile(rf"{atom_pos_type}\s*\n([\s\S]+)")
    block = block_pattern.search(contents).group()
    if block[-1] != "\n":
        block += "\n"
    atom_magnetism = []
    atom_symbol = []
    atom_block = []
    for i, symbol in enumerate(symbols):
        pattern = re.compile(rf"{symbol}\s*\n({_re_float})\s*\n(\d+)")
        sub_block = pattern.search(block)
        number = int(sub_block.group(2))

        # symbols, magnetism
        sym = [symbol] * number
        atom_mags = [float(sub_block.group(1))] * number
        for j in range(number):
            atom_symbol.append(sym[j])
            # atom_mass.append(masses[j])
            atom_magnetism.append(atom_mags[j])

        if i == ntype - 1:
            lines_pattern = re.compile(
                rf"{symbol}\s*\n{_re_float}\s*\n\d+\s*\n([\s\S]+)\s*\n"
            )
        else:
            lines_pattern = re.compile(
                rf"{symbol}\s*\n{_re_float}\s*\n\d+\s*\n([\s\S]+?)"
                + rf"\s*\n\w+\s*\n{_re_float}"
            )
        lines = lines_pattern.search(block)
        for j in [line.split() for line in lines.group(1).split("\n")]:
            atom_block.append(j)
    atom_block = np.array(atom_block)
    atom_magnetism = np.array(atom_magnetism)

    # position
    atom_positions = atom_block[:, 0:3].astype(float)

    def _get_index(labels, num):
        index = None
        res = []
        for label in labels:
            if label in atom_block:
                index = np.where(atom_block == label)[-1][0]
        if index is not None:
            res = atom_block[:, index + 1 : index + 1 + num].astype(float)

        return res, index

    # magnetism
    m_labels = ["mag", "magmom"]
    if "angle1" in atom_block or "angle2" in atom_block:
        import warnings

        warnings.warn(
            "Non-colinear angle-settings are not yet supported for this interface.",
            stacklevel=2,
        )
    mags, m_index = _get_index(m_labels, 1)
    try:  # non-colinear
        if m_index:
            atom_magnetism = atom_block[:, m_index + 1 : m_index + 4].astype(float)
    except IndexError:  # colinear
        if m_index:
            atom_magnetism = mags

    magnetic_moments = atom_magnetism if atom_magnetism.flatten().any() else None

    # to ase
    if atom_pos_type == "Direct":
        atoms = PhonopyAtoms(
            symbols=atom_symbol,
            cell=atom_lattice,
            scaled_positions=atom_positions,
            magnetic_moments=magnetic_moments,
        )
    elif atom_pos_type == "Cartesian":
        atoms = PhonopyAtoms(
            symbols=atom_symbol,
            cell=atom_lattice,
            positions=atom_positions * atom_lattice_scale,
            magnetic_moments=magnetic_moments,
        )
    else:
        warnings.warn("Only 'Direct' and 'Cartesian' are supported.", stacklevel=2)

    fd.close()
    return atoms, atom_potential, atom_basis, atom_offsite_basis


# READ ABACUS STRU -END-


#
# write ABACUS STRU
#


def write_abacus(filename, atoms, pps, orbitals=None, abfs=None):
    """Write structure to file."""
    with open(filename, "w") as f:
        f.write(get_abacus_structure(atoms, pps, orbitals, abfs))


def write_supercells_with_displacements(
    supercell,
    cells_with_displacements,
    ids,
    pps,
    orbitals=None,
    abfs=None,
    pre_filename="STRU",
    width=3,
):
    """Write supercells with displacements to files."""
    write_abacus("%s.in" % pre_filename, supercell, pps, orbitals, abfs)
    for i, cell in zip(ids, cells_with_displacements):
        filename = "{pre_filename}-{0:0{width}}".format(
            i, pre_filename=pre_filename, width=width
        )
        write_abacus(filename, cell, pps, orbitals, abfs)


def get_abacus_structure(atoms, pps, orbitals=None, abfs=None):
    """Return ABACUS structure in text."""
    empty_line = ""
    line = []
    line.append("ATOMIC_SPECIES")
    elements = list(Counter(atoms.symbols).keys())
    numbers = list(Counter(atoms.symbols).values())

    for _, elem in enumerate(elements):
        line.append(f"{elem}\t{atom_data[symbol_map[elem]][3]}\t{pps[elem]}")
    line.append(empty_line)

    if orbitals:
        line.append("NUMERICAL_ORBITAL")
        for _, elem in enumerate(elements):
            line.append(f"{orbitals[elem]}")
        line.append(empty_line)

    if abfs:
        line.append("ABFS_ORBITAL")
        for _, elem in enumerate(elements):
            line.append(f"{abfs[elem]}")
        line.append(empty_line)

    line.append("LATTICE_CONSTANT")
    line.append(str(1.0))
    line.append(empty_line)

    line.append("LATTICE_VECTORS")
    for i in range(3):
        line.append(" ".join(_list_elem2str(atoms.cell[i])))
    line.append(empty_line)

    line.append("ATOMIC_POSITIONS")
    line.append("Direct")
    line.append(empty_line)
    index = 0
    for i, elem in enumerate(elements):
        line.append(f"{elem}\n{0}\n{numbers[i]}")
        for j in range(index, index + numbers[i]):
            if atoms.magnetic_moments is not None:
                line_part = (
                    " ".join(_list_elem2str(atoms.scaled_positions[j])) + " 1 1 1"
                )
                # Add the magnetic moments part
                mag_mom = atoms.magnetic_moments[j]
                if isinstance(mag_mom, (list, np.ndarray)):
                    if (
                        len(mag_mom) == 3
                    ):  # Check the three components only if it is a list or numpy array
                        line_part += " mag " + " ".join(map(str, mag_mom))
                else:  # If single value (float), add directly
                    line_part += f" mag {mag_mom}"
                # Append the completed line
                line.append(line_part)
            else:
                line.append(
                    " ".join(_list_elem2str(atoms.scaled_positions[j])) + " " + "1 1 1"
                )
        line.append(empty_line)
        index += numbers[i]

    return "\n".join(line)


#
# set Force
#
def read_abacus_output(filename):
    """Read ABACUS forces from last self-consistency iteration."""
    force = None
    with open(filename, "r") as file:
        for line in file:
            if re.search(r"TOTAL ATOM NUMBER = [0-9]+", line):
                natom = int(re.search("[0-9]+", line).group())
            if re.search(r"TOTAL-FORCE \(eV/Angstrom\)", line):
                force = np.zeros((natom, 3))
                _match_pattern = r"^(\s*)([A-Za-z]*[0-9]+)((\s*[+-]?"
                _match_pattern += r"[0-9]+\.[0-9]+(e[+-][0-9]{2})?){3})(.*)$"
                _match = re.match(_match_pattern, line)
                while not _match:
                    line = file.readline()
                    _match = re.match(_match_pattern, line)
                iatom = 0
                while _match:
                    # print(_match.group(3).split())
                    fx, fy, fz = (_match.group(3).split()[i].strip() for i in range(3))
                    force[iatom] = (float(fx), float(fy), float(fz))
                    iatom += 1
                    if iatom == natom:
                        break
                    else:
                        line = file.readline()
                        _match = re.match(_match_pattern, line)

    if force is None:
        raise ValueError("Force data not found.")

    return force


def parse_set_of_forces(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    is_parsed = True
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))

        abacus_forces = read_abacus_output(filename)
        if check_forces(abacus_forces, num_atoms, filename, verbose=verbose):
            drift_force = get_drift_forces(
                abacus_forces, filename=filename, verbose=verbose
            )
            force_sets.append(np.array(abacus_forces) - drift_force)
        else:
            is_parsed = False

    if is_parsed:
        return force_sets
    else:
        return []


#
# tools
#


def _list_elem2str(a):
    """Convert type of list element to str."""

    def f_str(x):
        return f"{x:0<16.12f}"

    return list(map(f_str, a))



"""NaCl band structure calculation example."""

import phonopy
from phonopy.phonon.band_structure import get_band_qpoints

phonon = phonopy.load(
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    unitcell_filename="POSCAR-unitcell",
    force_sets_filename="FORCE_SETS",
    born_filename="BORN",
)
points = get_band_qpoints(
    [[[0.5, 0, 0.5], [0, 0, 0], [0.5, 0.5, 0.5], [0.5, 0.25, 0.75]]], 51
)
phonon.run_band_structure(points, labels=["X", r"$\Gamma$", "L", "W"])
phonon.plot_band_structure().show()



"""Tests for phonon calculation at specific q-points."""

import numpy as np

from phonopy import Phonopy
from phonopy.physical_units import get_physical_units


def test_Qpoints(ph_nacl_nofcsym: Phonopy):
    """Test phonon calculation at specific q-points by NaCl."""
    phonon = ph_nacl_nofcsym
    qpoints = [[0, 0, 0], [0, 0, 0.5]]
    phonon.run_qpoints(qpoints, with_dynamical_matrices=True)
    for i, _ in enumerate(qpoints):
        dm = phonon.qpoints.dynamical_matrices[i]
        dm_eigs = np.linalg.eigvalsh(dm).real
        eigs = phonon.qpoints.eigenvalues[i]
        freqs = phonon.qpoints.frequencies[i] / get_physical_units().DefaultToTHz
        np.testing.assert_allclose(dm_eigs, eigs)
        np.testing.assert_allclose(freqs**2 * np.sign(freqs), eigs)


def test_Qpoints_with_NAC_qdirection(ph_nacl: Phonopy):
    """Test phonon calculation at specific q-points by NaCl."""
    phonon = ph_nacl
    qpoints = [[0, 0, 0]]
    phonon.run_qpoints(qpoints)
    freqs = phonon.get_qpoints_dict()["frequencies"]
    np.testing.assert_allclose(
        freqs, [[0, 0, 0, 4.61643516, 4.61643516, 4.61643516]], atol=1e-5
    )
    phonon.run_qpoints(qpoints, nac_q_direction=[1, 0, 0])
    freqs = phonon.get_qpoints_dict()["frequencies"]
    np.testing.assert_allclose(
        freqs, [[0, 0, 0, 4.61643516, 4.61643516, 7.39632718]], atol=1e-5
    )



"""Tests for TetrahedronMesh."""

import os
from io import StringIO

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.file_IO import parse_FORCE_SETS
from phonopy.interface.vasp import read_vasp
from phonopy.phonon.tetrahedron_mesh import TetrahedronMesh

data_dir = os.path.dirname(os.path.abspath(__file__))

dos_str = """-0.672024 0.000000 0.029844 0.005522 0.731712 0.029450 1.433580 0.116998
2.135448 0.200612 2.837315 0.329781 3.539183 0.905737 4.241051 1.808068
4.942919 1.921464 5.644787 1.415087 6.346654 1.281276 7.048522 1.127877
7.750390 0.725551 8.452258 1.611979 9.154126 1.157525 9.855993 0.929781
10.557861 0.705942 11.259729 0.630369 11.961597 0.557250 12.663465 0.487940
13.365332 0.143148 14.067200 0.231518 14.769068 0.205816 15.470936 0.374575
16.172804 0.781758 16.874671 0.352501 17.576539 0.473537 18.278407 0.310505
18.980275 0.184665 19.682143 0.018812 20.384010 0.023455 21.085878 0.028958
21.787746 0.035963 22.489614 0.047177 23.191481 0.068537 23.893349 0.092864
24.595217 0.204422 25.297085 0.252177 25.998953 0.676881 26.700820 0.000000"""


@pytest.mark.parametrize("langs", [("C", "C"), ("Py", "C"), ("C", "Py")])
def test_Amm2(langs):
    """Test of DOS calculation using TetrahedronMesh for Amm2 crystal."""
    data = np.loadtxt(StringIO(dos_str))
    phonon = _get_phonon("Amm2", [3, 2, 2], [[1, 0, 0], [0, 0.5, -0.5], [0, 0.5, 0.5]])
    mesh = [11, 11, 11]
    primitive = phonon.primitive
    phonon.run_mesh([11, 11, 11])
    weights = phonon.mesh.weights
    frequencies = phonon.mesh.frequencies
    grid_address = phonon.mesh.grid_address
    ir_grid_points = phonon.mesh.ir_grid_points
    grid_mapping_table = phonon.mesh.grid_mapping_table
    th_mesh = TetrahedronMesh(
        primitive,
        frequencies,
        mesh,
        np.array(grid_address, dtype="int64"),
        np.array(grid_mapping_table, dtype="int64"),
        ir_grid_points,
        lang=langs[0],
    )
    th_mesh.set(value="I", division_number=40, lang=langs[1])
    freq_points = th_mesh.get_frequency_points()
    dos = np.zeros_like(freq_points)

    for i, iw in enumerate(th_mesh):
        dos += np.sum(iw * weights[i], axis=1)

    dos_comp = np.transpose([freq_points, dos]).reshape(10, 8)
    np.testing.assert_allclose(dos_comp, data, atol=1e-5)


def _show(freq_points, dos):
    data = []
    for f, d in zip(freq_points, dos):
        data.append([f, d])
    data = np.reshape(data, (10, 8))
    for row in data:
        print(("%f " * 8) % tuple(row))


def _get_phonon(spgtype, dim, pmat):
    cell = read_vasp(os.path.join(data_dir, "POSCAR_%s" % spgtype))
    phonon = Phonopy(cell, np.diag(dim), primitive_matrix=pmat)
    force_sets = parse_FORCE_SETS(
        filename=os.path.join(data_dir, "FORCE_SETS_%s" % spgtype)
    )
    phonon.dataset = force_sets
    phonon.produce_force_constants()
    return phonon
